#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from frm.term_structures.zero_curve import * # noqa: F403
from frm.term_structures.zero_curve_helpers import * # noqa: F403
from frm.term_structures.historical_swap_index_fixings import * # noqa: F403
from frm.term_structures.swap_curve import * # noqa: F403




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from frm.utils.business_day_calendar import * # noqa: F403
from frm.utils.daycount import * # noqa: F403
from frm.utils.interpolation import * # noqa: F403
from frm.utils.schedule import * # noqa: F403
from frm.utils.tenor import * # noqa: F403
from frm.utils.utilities import * # noqa: F403
from frm.utils.settings import * # noqa: F403
from frm.utils.custom_errors_and_warnings import * # noqa: F403


# -*- coding: utf-8 -*-
import os
import numpy as np
import scipy.fft
import scipy
from scipy.stats import norm
from numba import njit
from typing import Tuple
import warnings

from frm.pricing_engine.garman_kohlhagen import garman_kohlhagen_solve_implied_vol, garman_kohlhagen_solve_strike_from_delta
from frm.pricing_engine.monte_carlo_generic import normal_corr
from frm.pricing_engine.cosine_method_generic import get_cos_truncation_range

if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))

VALID_HESTON_PRICING_METHODS = [
    'heston_1993',
    'heston_carr_madan_gauss_kronrod_quadrature',
    'heston_carr_madan_fft_w_simpsons',
    'heston_cosine',
    'heston_lipton'
]

def heston_price_vanilla_european(S0, tau, r, q, cp, K, var0, vv, kappa, theta, rho, lambda_, pricing_method):
    match pricing_method:
        case 'heston_cosine':
            return heston_cosine_price_vanilla_european(S0=S0, tau=tau, r=r, q=q, cp=cp, K=K, var0=var0, vv=vv,
                                                        kappa=kappa, theta=theta, rho=rho)
        case 'heston_1993':
            return heston1993_price_vanilla_european(S0=S0, tau=tau, r=r, q=q, cp=cp, K=K, var0=var0, vv=vv,
                                                     kappa=kappa, theta=theta, rho=rho, lambda_=lambda_)
        case 'heston_carr_madan_gauss_kronrod_quadrature':
            return heston_carr_madan_price_vanilla_european(S0=S0, tau=tau, r=r, q=q, cp=cp, K=K, var0=var0, vv=vv,
                                                            kappa=kappa, theta=theta, rho=rho, integration_method=0)
        case 'heston_carr_madan_fft_w_simpsons':
            return heston_carr_madan_price_vanilla_european(S0=S0, tau=tau, r=r, q=q, cp=cp, K=K, var0=var0, vv=vv,
                                                            kappa=kappa, theta=theta, rho=rho, integration_method=1)
        case 'heston_lipton':
            return heston_lipton_price_vanilla_european(S0=S0, tau=tau, r=r, q=q, cp=cp, K=K, var0=var0, vv=vv,
                                                        kappa=kappa, theta=theta, rho=rho)
        case _:
            raise ValueError("Invalid 'pricing_method:", pricing_method)




def validate_input(var, var_name, validation_fn):
    if np.isscalar(var):
        if not validation_fn(var):
            raise ValueError(f"'{var_name}' has invalid values: {var}")
    else:
        if not np.all(validation_fn(var)):
            raise ValueError(f"'{var_name}' has invalid values: {var}")


def heston_calibrate_vanilla_smile(
        volatility_quotes: np.array,
        delta_of_quotes: np.array,
        S0: float,
        r: float,
        q: float,
        tau: float,
        cp: np.array,
        delta_convention: str = None,
        pricing_method='heston_cosine') -> Tuple[float, float, float, float, float, float, np.array, float]:
    """
    Fit the Heston model to the FX market implied volatility smile.

    Parameters:
    - volatility_quotes (np.array): Vector of volatilities (i.e. implied volatility quotes) to calibrate to.
    - delta_of_quotes (np.array): Vector of delta values of the volatility quotes
    - S0 (float): Initial asset price (specified in # of units of domestic per 1 unit of foreign currency for Garman-Kohlhagen)
    - r (float): Risk-free interest rate (Domestic interest rate for Garman-Kohlhagen)
    - q (float): Interest rate of the asset (Foreign interest rate for Garman-Kohlhagen)
    - tau (float): Time to maturity in years
    - cp (np.array): Vector of option types (1 for call, -1 for put)
    - delta_convention (str): Delta convention ('prem-adj' or 'prem-adj-fwd')
    - pricing_method (str): Pricing method for the Heston model. Default is 'carr_madan_gauss_kronrod_quadrature'

    Returns:
    - Tuple: Initial variance (var0),
             vol of vol (vv),
             Mean reversion speed to the long-run variance (kappa),
             long-run mean variance (theta),
             market price of volatility risk (lambda_),
             correlation (rho),
             vector of implied volatilities (IV),
             sum of squared errors (SSE)

    References:
    [1] Janek, A., Kluge, T., Weron, R., Wystup, U. (2010). "FX smile in the Heston model"

    The Heston SDE is:
        dS(t) = µ*S*dt + σ(t)*S*dW1(t)
        dσ(t) = kappa(theta - σ(t))*dt + vv*σ(t)*dW2(t)

    The Heston model is defined by six parameters,
    - var0: Initial variance.
    - vv: Volatility of volatility.
    - kappa: Mean reversion speed to the long-run variance
    - theta: Long-run variance.
    - lambda_: Market price of volatility risk.
    - rho: Correlation (between stock price and volatility).

    The function initially estimates strikes using the Garman-Kohlhagen model.
    It then optimizes the Heston parameters to minimize the sum of squared errors between market and model-implied volatilises.
    """

    def calibration_helper(param, var0, kappa, S0, tau, r, q, cp, K, volatility_quotes, method):
        """
        Compute the sum of squared errors (SSE) between market and model implied volatilities.

        Parameters:
        param (list): [vol of vol (vv), long-run variance (theta), correlation (rho)]
        var0 (float): Initial variance.
        kappa (float): Mean reversion speed to the long-run variance
        S0 (float): Initial asset price (specified in # of units of domestic per 1 unit of foreign currency for Garman-Kohlhagen)
        tau (float): Time to maturity in years.
        r (float): Risk-free interest rate (Domestic interest rate for Garman-Kohlhagen)
        q (float): Interest rate of the asset (Foreign interest rate for Garman-Kohlhagen)
        cp (array): Option types; 1 for call, -1 for put.
        K (array): Vector of strike prices.
        volatility_quotes (array): Vector of market implied volatilises.
        method (int): 0 for SSE, 1 for P and IV.

        Returns:
        float: Sum of squared errors between model and market implied volatilities.
        """
        vv, theta, rho = param
        if vv < 0.0 or theta < 0.0 or abs(rho) > 1.0:
            # warnings.warn("Invalid value for vv, theta or rho encountered")
            return np.inf

        P = np.zeros(nb_strikes)
        IV = np.zeros(nb_strikes)

        if pricing_method == 'heston_cosine':
            P = heston_cosine_price_vanilla_european(S0=S0, tau=tau, r=r, q=q, cp=cp, K=strikes, var0=var0,
                                                     vv=vv, kappa=kappa, theta=theta, rho=rho)
        else:
            # Integral required for each strike hence can't be vectorised
            for i in range(nb_strikes):
                P[i] = heston_price_vanilla_european(S0=S0, tau=tau, r=r, q=q, cp=cp[i], K=K[i], var0=var0, vv=vv, kappa=kappa,
                                            theta=theta, rho=rho, lambda_=lambda_, pricing_method=pricing_method)

        for i in range(nb_strikes):
            if P[i] < 0.0:
                IV[i] = -1.0
            else:
                IV[i] = garman_kohlhagen_solve_implied_vol(S0=S0, tau=tau, r_d=r, r_f=q, cp=cp[i], K=strikes[i], X=P[i], vol_guess=volatility_quotes[i])

        SSE = np.sum((volatility_quotes - IV)**2)

        if method == 0:
            return SSE
        elif method == 1:
            return P, IV, SSE

    # Input validation
    validate_input(S0, 'S0', lambda x: x > 0 and not np.isnan(x))
    validate_input(tau, 'tau', lambda x: x > 0 and not np.isnan(x))
    validate_input(r, 'r', lambda x: not np.isnan(x))
    validate_input(q, 'q', lambda x: not np.isnan(x))
    validate_input(cp, 'cp', lambda x: np.isin(x, [-1, 1]))
    if pricing_method not in VALID_HESTON_PRICING_METHODS:
        raise ValueError(f"'pricing_method' is invalid: {pricing_method}")

    # Calculate strikes for market deltas
    strikes = garman_kohlhagen_solve_strike_from_delta(S0=S0,tau=tau,r_d=r,r_f=q,vol=volatility_quotes,signed_delta=delta_of_quotes,delta_convention=delta_convention)
    nb_strikes = len(strikes)

    # Set the initial volatility, var0, to the implied ATM market volatility
    # var0 will NOT be solved in the calibration
    var0 = np.power(volatility_quotes[nb_strikes // 2], 2)

    # Set mean reversion to the long-run variance to constant value = 1.5
    # The influence of mean reversion can be compensated by the volatility of variance
    # kappa could be adjusted to enforce a valid Feller condition but this may detrimentally impact the fit to the market
    # The Feller condition is often required to be violated in order to get a good fit to market data
    kappa = 1.5

    # lambda_ is the market price of volatility risk
    # lambda_ will NOT be solved in the calibration
    # We set it to zero, given we are calibrating off option prices which already contain this feature
    # If calibrating of historical data, this field would need to be considered.
    lambda_ = 0

    # Set initial values for vv, theta, rho (the parameters we are solving for)
    init_param = np.array([2 * np.sqrt(var0), 2*var0, 0])

    res = scipy.optimize.minimize(
        lambda param: calibration_helper(param, var0=var0, kappa=kappa, S0=S0, tau=tau, r=r, q=q, cp=cp, K=strikes, volatility_quotes=volatility_quotes, method=0),
        init_param)
    vv, theta, rho = res.x

    if 2 * kappa * theta - vv**2 <= 0.0:
        # In the Heston model, the Feller condition is often required to be violated in order to get a good fit to market data
        # warnings.warn("Feller condition violated.")
        pass

    P, IV, SSE = calibration_helper(param=[vv, theta, rho], var0=var0, kappa=kappa, S0=S0, tau=tau, r=r, q=q, cp=cp, K=strikes, volatility_quotes=volatility_quotes, method=1)

    if SSE == np.inf:
        warnings.warn("Calibration failed, SSE==inf")
        return None
    else:
        return var0, vv, kappa, theta, rho, lambda_, IV, SSE


@njit(fastmath=True, cache=True)
def heston_1993_vanilla_european_integral(φ, m, S0, K, tau, r, q, var0, vv, kappa, theta, rho, lambda_=0):
    """
    Defines the integral for pricing an Vanilla European option per the analytic Heston 1993 formula
    This implementation uses the 2nd form of the Heston Characteristic function, detailed in Albrecher 2006 is used as it is more numerically stable.
    This in an auxiliary function for heston1993_price_fx_vanilla_european(), separated so @jit can be used

    Parameters:
    φ (complex): Point at which the auxiliary function is evaluated.
    m (int): Index for specific calculation (either 1 or 2).
    S0 (float): Initial asset price (specified in # of units of domestic per 1 unit of foreign currency for Garman-Kohlhagen)
    K (float): Strike price.
    var0 (float): Initial variance.
    vv (float): Volatility of volatility.
    r (float): Risk-free interest rate (Domestic interest rate for Garman-Kohlhagen)
    q (float): Interest rate of the asset (Foreign interest rate for Garman-Kohlhagen)
    tau (float): Time to maturity in years.
    kappa (float): Mean reversion speed to the long-run variance
    theta (float): Long-run variance
    rho (float): Correlation between asset price and asset volatility Wiener processes.
    lambda_ (float), optional: Market price of volatility risk. Set to zero if calibration was off market option prices.

    Returns:
    float: Value of the auxiliary function at point `φ`.

    References:
    [1] S.Heston, (1993) A Closed-Form Solution for Options with Stochastic Volatility with Applications to Bond and Currency Options
    [2] H.Albrecher, P.Mayer, W.Schoutens, J.Tistaert (2006) The little Heston trap, Wilmott Magazine, January: 83–92.
    [3] A.Janek, T.Kluge, R.Weron, U.Wystup (2010) FX smile in the Heston model.
    """

    mu = r - q

    # x per equation (11) from Heston, 1993
    x = np.log(S0)
    # a, u, b per equation (12) from Heston, 1993
    a = kappa * theta
    u = [0.5, -0.5]
    b = [kappa + lambda_ - rho * vv, kappa + lambda_]

    σ = vv # In Heston, 1993, the volatility of the volatility is defined as σ

    # d is per per equation (17) from Heston, 1993
    d = np.sqrt((rho *σ*φ*1j - b[m-1])**2 - σ**2 * (2*u[m-1]*φ*1j - φ**2))

    # There are two formulas for "g", in the Heston characteristic function because d, an input to g, has two roots.
    # Let "g1", be the presentation in the original Heston 1993 paper (on page 5 of 17 of [1])
    # Let "g2", be the alternative, that is detailed in Albrecher 2006, "The Little Heston Trap".
    # g2 = 1 / g1.
    # We use g2, as using g1 leads to numerical problems based on most software implementations of calculation involving complex numbers
    g2 = (b[m-1] - rho * σ * φ * 1j - d) \
        / (b[m-1] - rho * σ * φ * 1j + d)

    # C, D and characteristic function f, per equation (17) from Heston, 1993
    C = mu * φ * 1j * tau + a / (vv ** 2) * ((b[m-1] - rho * σ * φ * 1j - d) * tau - 2 * np.log((1 - g2 * np.exp(-d * tau)) / (1 - g2)))
    D = (b[m-1] - rho * σ * φ * 1j - d) / (vv ** 2) * ((1 - np.exp(-d * tau)) / (1 - g2 * np.exp(-d * tau)))
    chf = np.exp(C + D * var0 + 1j * φ * x)

    # Function inside the integral in equation (18) from Heston, 1993
    F = np.real(np.exp(-1j * φ * np.log(K)) * chf / (1j * φ))
    return F


def heston1993_price_vanilla_european(S0, tau, r, q, cp, K, var0, vv, kappa, theta, rho, lambda_=0):
    """
    Calculate the price of a European Vanilla option using the analytical Heston 1993 formulae
    The 2nd form of the Heston Characteristic function, detailed in Albrecher 2006 is used as it is more numerically stable.

    Parameters:
    S0 (float): Initial asset price (specified in # of units of domestic per 1 unit of foreign currency for Garman-Kohlhagen)
    tau (float): Time to expiry in years.
    r (float): Risk-free interest rate (Domestic interest rate for Garman-Kohlhagen)
    q (float): Interest rate of the asset (Foreign interest rate for Garman-Kohlhagen).
    cp (int): Call (1) or Put (-1) option.
    K (float): Strike price.
    var0 (float): Initial variance.
    vv (float): Volatility of volatility.
    kappa (float): Mean reversion speed to the long-run variance
    theta (float): Long-run variance.
    rho (float): Correlation between asset price and asset volatility Wiener processes.
    lambda_ (float), optional: Market price of volatility risk. Set to zero if calibration was off market option prices.

    Returns:
    float: Option price.

    References:
    [1] S.Heston, (1993) A Closed-Form Solution for Options with Stochastic Volatility with Applications to Bond and Currency Options
    [2] H.Albrecher, P.Mayer, W.Schoutens, J.Tistaert (2006) The little Heston trap, Wilmott Magazine, January: 83–92.
    [3] A.Janek, T.Kluge, R.Weron, U.Wystup (2010) FX smile in the Heston model.

    Written by Agnieszka Janek and Rafal Weron (2010.07.07)
    Revised by Rafal Weron (2010.10.08, 2010.12.27)
    """

    # Equation (18) from Heston, 1993
    P1 = 0.5 + 1/np.pi * scipy.integrate.quad(func=lambda φ: heston_1993_vanilla_european_integral(φ=φ, m=1, S0=S0, K=K, tau=tau, r=r, q=q, var0=var0, vv=vv, kappa=kappa, theta=theta, rho=rho, lambda_=lambda_), a=0, b=np.inf, epsrel=1e-8)[0]
    P2 = 0.5 + 1/np.pi * scipy.integrate.quad(func=lambda φ: heston_1993_vanilla_european_integral(φ=φ, m=2, S0=S0, K=K, tau=tau, r=r, q=q, var0=var0, vv=vv, kappa=kappa, theta=theta, rho=rho, lambda_=lambda_), a=0, b=np.inf, epsrel=1e-8)[0]

    Pplus = (1 - cp) / 2 + cp * P1   # Pplus = N(d1)
    Pminus = (1 - cp) / 2 + cp * P2  # Pminus = N(d2)

    # tk is this for spot or forward Δ, matlab code assumed spot delta input?
    X = cp * (S0 * np.exp(-q * tau) * Pplus - K * np.exp(-r * tau) * Pminus)

    return X



def heston_carr_madan_price_vanilla_european(S0, tau, r, q, cp, K, var0, vv, kappa, theta, rho, integration_method=0):
    """
     Calculate European option price using the Heston model via Carr-Madan approach.

     Parameters:
     kappa, theta, vv, rho, var0 (float): Heston parameters.
     integration_method (int, optional): 0 for Gauss-Kronrod quadrature, 1 for FFT + Simpson's rule. Default is 0.

     Returns:
     float: Option price.

     References:
     [1] Albrecher et al. (2006) "The little Heston trap."
     [2] Carr, Madan (1998) "Option valuation using the Fast Fourier transform."
     [3] Janek et al. (2010) "FX smile in the Heston model."
     [4] Schmelzle (2010) "Option Pricing Formulae using Fourier Transform."

     Authors:
     Written by Agnieszka Janek (2010.07.23)
     Revised by Rafal Weron (2010.10.08)
     Revised by Agnieszka Janek and Rafal Weron (2010.10.21, 2010.12.27)
    """

    if cp == 1:
        alpha = 0.75
    elif cp == -1:
        alpha = 1.75
    else:
        raise ValueError

    log_S0 = np.log(S0)
    log_K = np.log(K)

    if integration_method == 0:
        # Integrate using adaptive Gauss-Kronrod quadrature
        args = (cp, log_S0, log_K, tau, r, q, var0, vv, kappa, theta, rho, alpha)
        result, _ = scipy.integrate.quad(func=heston_fft_vanilla_european_integral, a=0, b=np.inf, args=args)
        y = np.exp(-cp * log_K * alpha) * result / np.pi
    elif integration_method == 1:
        # Fast Fourier Transform with Simpson's rule (as suggested in [2])
        N = 2**10
        eta = 0.25
        v = np.arange(0, N) * eta

        lambda_ = 2*np.pi/(N*eta)
        b = N*lambda_/2 # Equation (20) per Carr-Madan, 1998
        ku = -b + lambda_ * np.array(range(0,N)) # Equation (19) per Carr-Madan, 1998

        u = v - (cp * alpha + 1) * 1j
        chf = chf_heston_albrecher2007(u=u, log_S0=log_S0, tau=tau, r=r, q=q, var0=var0, vv=vv, kappa=kappa, theta=theta, rho=rho)
        F = chf * np.exp(-r * tau) / (alpha**2 + cp * alpha - v**2 + 1j * (cp * 2 * alpha + 1) * v)

        # Use Simpson's approximation to calculate FFT (see [2])
        simpson_weights = get_simpson_weights(N)
        fft_func = np.exp(1j * b * v) * F * eta * simpson_weights
        payoff = np.real(scipy.fft.fft(fft_func))
        option_value = np.exp(-cp * ku * alpha) * payoff / np.pi

        y = np.interp(log_K, ku, option_value)

    return y


def get_simpson_weights(n):
    # 1/3, then alternating 2/3, 4/3. i.e 1/3, 2/3, 4/3, 2/3, 4/3, 2/3, 4/3, ....
    weights = np.array([1] + [4, 2] * ((n-2) // 2) + [1])
    return weights / 3


@njit(fastmath=True, cache=True)
def heston_fft_vanilla_european_integral(v, cp, log_S0, log_K, tau, r, q, var0, vv, kappa, theta, rho, alpha):
    """
    Auxiliary function for heston_carr_madan_vanilla_european.

    Parameters:
        v: Evaluation points for auxiliary function.
        cp: Option type (1 for call, -1 for put).
        log_S0 (float): Natural Log of the initial asset price (in # of domestic currency units per 1 foreign currency unit for Garman-Kohlhagen)
        log_K: Natural log of strike price (in # of domestic currency units per 1 foreign currency unit)
        tau (float): Time to maturity (in years)
        r (float): Risk-free interest rate (Domestic interest rate for Garman-Kohlhagen)
        q (float): Interest rate of the asset (Foreign interest rate for Garman-Kohlhagen)
        var0 (float): Initial variance.
        vv (float): Volatility of volatility.
        kappa (float): Mean reversion speed to the long-run variance
        theta (float): Long-run variance.
        rho (float): Correlation between asset price and asset volatility Wiener processes.
        alpha: Damping coefficient.

    Returns:
        The values of the auxiliary function evaluated at points v.
    """

    u = v - (cp * alpha + 1) * 1j
    chf = chf_heston_albrecher2007(u=u, log_S0=log_S0, tau=tau, r=r, q=q, var0=var0, vv=vv, kappa=kappa, theta=theta, rho=rho)
    ftt_func = chf * np.exp(-r * tau) / (alpha**2 + cp * alpha - v ** 2 + 1j * (cp * 2 * alpha + 1) * v)
    return np.real(np.exp(-1j * v * log_K) * ftt_func)


@njit(fastmath=True, cache=True)
def chf_heston_albrecher2007(u, log_S0, tau, r, q, var0, vv, kappa, theta, rho):
    """
    Compute the characteristic function for the Heston model.

    Parameters:
    u (float or np.array): The argument of the characteristic function.
    log_S0 (float): Natural Log of the initial spot price (in # of domestic currency units per 1 foreign currency unit)
    tau (float): Time to maturity (in years)
    r (float): Risk-free interest rate (Domestic interest rate for Garman-Kohlhagen)
    q (float): Interest rate of the asset (Foreign interest rate for Garman-Kohlhagen)
    var0 (float): Initial variance.
    vv (float): Volatility of volatility.
    kappa (float): Mean reversion speed to the long-run variance
    theta (float): Long-run variance
    rho (float): Correlation between asset price and asset volatility Wiener processes.

    Returns:
        charFunc (float or np.array): Value of the characteristic function at u.

    [1] S.Heston, (1993) A Closed-Form Solution for Options with Stochastic Volatility with Applications to Bond and Currency Options
    [2] Albrecher, Hansjoerg & Mayer, Philipp & Schoutens, Wim & Tistaert, Jurgen. (2007). The Little Heston Trap. Wilmott. 83-92

    """

    # The symbol use in heston model academic literature is inconsistent
    # Remap the parameters per the symbols used in [2] so comparison to the formulae on page 4 of 21 is easier
    k = kappa # mean reversion rate of the variance to the long run variance
    η = theta # the long run variance
    λ = vv # volatility of the variance
    σ0 = np.sqrt(var0) # initial volatility
    r = r
    q = q

    # Per definition in equation (1) of [2] on page 4/21
    d = np.sqrt((rho*λ*u*1j - k)**2 + λ**2 * (1j * u + u ** 2))

    # There are two formulas for "g", in the Heston characteristic function because d, an input to g, has two roots.
    # Let "g1", be the presentation in the original Heston 1993 paper (on page 5 of 17 of [1])
    # Let "g2", be the alternative, that is detailed in [2], "The little heston trap".
    # g2 = 1 / g1.
    # We use g2, as using g1 leads to numerical problems based on most software implementations of calculation involving complex numbers
    g2 = (k - rho * λ * 1j * u - d) \
      / (k - rho * λ * 1j * u + d)

    # 1st inner exponential term on line 1 of 3, in equation (2) in [2]
    A = 1j * u * (log_S0 + (r - q) * tau)

    # 2nd inner exponential term on line 2 of 3, in equation (2) in [2]
    B = η*k*(λ**-2) * ((k - rho*λ*1j*u - d) * tau - 2 * np.log((1 - g2*np.exp(-d*tau)) / (1 - g2)))

    # 3rd inner exponential term on line 3 of 3, in equation (2) in [2]
    C = (σ0**2)*(λ**-2) * (k - rho*λ*1j*u - d) * (1 - np.exp(-d*tau)) / (1 - g2*np.exp(-d*tau))

    return np.exp(A + B + C)


# Note this function (chf_heston_fang2008) is NOT speed up by use of numba/jit
@njit(fastmath=True, cache=True)
def chf_heston_fang2008(u, tau, r, q, var0, vv, kappa, theta, rho):
    """
    Compute the characteristic function for the Heston model per Fang 2008

    Parameters:
    u (float or np.array): The argument values of the characteristic function.
    tau (float): Time to maturity (in years)
    r (float): Risk-free interest rate (Domestic interest rate for Garman-Kohlhagen)
     q (float): Interest rate of the asset (Foreign interest rate for Garman-Kohlhagen)
    var0 (float): Initial variance.
    vv (float): Volatility of volatility.
    kappa (float): Mean reversion speed to the long-run variance
    theta (float): Long-run variance
    rho (float): Correlation between asset price and asset volatility Wiener processes.

    Returns:
        charFunc (float or np.array): Value of the characteristic function at u.

    References
    [1] Fang, Fang & Oosterlee, Cornelis. (2008). A Novel Pricing Method for European Options Based on Fourier-Cosine Series Expansions. SIAM J. Scientific Computing. 31. 826-848. 10.1137/080718061.
    [2] Albrecher, Hansjoerg & Mayer, Philipp & Schoutens, Wim & Tistaert, Jurgen. (2007). The Little Heston Trap. Wilmott. 83-92
    """

    mu = r - q # mu = r - q

    # Map to the symbols used in [1] so for easier comparison to the paper
    λ = kappa  # mean reversion speed to the long-run variance
    u_bar = theta # long-run variance
    η = vv # volatility of the volatility
    u0 = var0 # initial variance

    ω = u
    W = λ - 1j*rho*η*ω # helper term to make simpler code (not in [3])
    D = np.sqrt( W**2 + (ω**2 + 1j*ω) * (η**2))
    G2 = (W - D) / (W + D) # G2 is more stable per [2]
    exp_D_tau = np.exp(-D*tau)

    inner_exp_1 = 1j*ω*mu*tau + (u0/(η**2)) * ((1-exp_D_tau)/(1 - G2*exp_D_tau)) * (W-D)
    inner_exp_2 = (λ*u_bar)/(η**2) * ((tau * (W - D)) - 2*np.log( (1-G2*exp_D_tau) / (1-G2) ))

    chf = np.exp(inner_exp_1 + inner_exp_2)

    return chf


def calculate_Uk_european_options(cp, a, b, k):
    """
    Compute Uk for valuing vanilla european call or put options under Heston using COS method.

    Parameters:
    - cp (int): 1 for call, -1 for put
    - a (float): Lower truncation boundary
    - b (float): Upper truncation boundary
    - k (np.array): Array of k values, [0, 1, 2, .... to N] where N is the number of expansion terms

    Returns:
    - np.array: Coefficients

    References:
    [1] Fang, Fang & Oosterlee, Cornelis. (2008). A Novel Pricing Method for European Options Based on Fourier-Cosine Series Expansions. SIAM J. Scientific Computing. 31. 826-848. 10.1137/080718061.
    """

    def calc_chi(k, c, d):
        # Calculate χ, chi, Per equation 22 from [1] (page 6 of 21)
        χ = np.zeros(shape=k.shape)
        term1 = 1.0 / (1.0 + np.power((k * np.pi / (b - a)) , 2.0))
        term2 = np.cos(k * np.pi * (d - a)/(b - a)) * np.exp(d)
        term3 = np.cos(k * np.pi  * (c - a) / (b - a)) * np.exp(c)
        term4 = k * np.pi / (b - a) * np.sin(k * np.pi * (d - a) / (b - a))
        term5 = k * np.pi / (b - a) * np.sin(k * np.pi * (c - a) / (b - a)) * np.exp(c)
        χ = term1 * (term2 - term3 + term4 - term5)
        return χ

    def calc_psi(k, c, d):
        # Calculate Ψ, psi, per equation 23 from [1] (page 6 of 21)
        Ψ = np.zeros(shape=k.shape)
        Ψ[1:] = (np.sin(k[1:] * np.pi * (d-a) / (b-a)) - np.sin(k[1:] * np.pi * (c-a) / (b-a))) * (b-a) / (k[1:] * np.pi)
        Ψ[0] = d - c
        return Ψ

    if cp == 1:
        # For call, c=0, d=b per equation 29 from [1] (page 7 of 21)
        χ = calc_chi(k=k, c=0, d=b)
        Ψ = calc_psi(k=k, c=0, d=b)
        Uk_call = (2 / (b-a)) * (χ - Ψ)
        return Uk_call
    elif cp == -1:
        # For put, c=a, d]0, per equation 29 from [1] (page 7 of 21)
        χ = calc_chi(k=k, c=a, d=0)
        Ψ = calc_psi(k=k, c=a, d=0)
        Uk_put = (2 / (b-a)) * (-χ + Ψ)
        return Uk_put


def heston_cosine_price_vanilla_european(S0, tau, r, q, cp, K, var0, vv, kappa, theta, rho, N=160, L=10, calculate_via_put_call_parity=True):
    """
    Computes the call or put option prices using the COS method.

    Parameters:
    S0 (float): Initial asset price
    tau (float): time to expiry
    r (float): Risk-free interest rate (Domestic interest rate for Garman-Kohlhagen)
    q (float): Interest rate of the asset (Foreign interest rate for Garman-Kohlhagen)
    cp (int): 1 for call and -1 for put
    K (list or np.array): List of strike prices
    var0 (float): Initial variance
    vv (float): Volatility of volatility
    kappa (float): Mean reversion speed to the long-run variance
    theta (float): Long-run variance
    rho (float): Correlation between asset price and asset volatility Wiener processes.
    N (int): Number of expansion terms (<160 should be sufficient per Fang, 2008)
    L (float): Size of truncation domain (for N=160, set L=10 per Fang, 2008)
    calculate_via_put_call_parity (boolean)

    Returns:
    - np.array: Option prices

    References:
    [1] Fang, Fang & Oosterlee, Cornelis. (2008). A Novel Pricing Method for European Options Based on Fourier-Cosine Series Expansions. SIAM J. Scientific Computing. 31. 826-848. 10.1137/080718061.
    """

    cp = np.atleast_1d(cp).astype(float)
    K = np.atleast_1d(K).astype(float)
    assert cp.shape == K.shape

    result = np.full(K.shape, np.nan)

    x0 = np.log(S0 / K) # Per [1] in section 3.1 on page 6 of 21

    if True:
        # Apply the truncation method per appendix 11 of [1]
        mu = r - q # technically we should use the drift implied from market FX forward rate, but the IR parity forward rate should be close enough
        model_param =  {'tau': tau,
                        'mu': mu,
                        'var0': var0,
                        'vv': vv,
                        'kappa': kappa,
                        'theta': theta,
                        'rho': rho}
        a, b = get_cos_truncation_range(model='heston', L=L, model_param=model_param)
    else:
        # This is a simpler truncation method
        # It produces a wider bound than the more complex method detailed in [1]
        a, b = -L * np.sqrt(tau), L * np.sqrt(tau)

    # Summation from k = 0 to k = N-1
    k = np.arange(N)

    u = (k*np.pi)/(b-a)
    chf = chf_heston_fang2008(u=u, tau=tau, r=r, q=q, var0=var0, vv=vv, kappa=kappa, theta=theta, rho=rho)
    Fk = np.real(chf[:, np.newaxis]  * np.exp(1j * k[:, np.newaxis] * np.pi * (x0 - a)/(b-a)))
    Fk[0] = 0.5 * Fk[0] # Per page 3/21 of [1], "where Σ′ indicates that the first term in the summation is weighted by one-half"


    if calculate_via_put_call_parity or (cp == -1).any():
        Uk_put = calculate_Uk_european_options(cp=-1, a=a, b=b, k=k)[:, np.newaxis] # Per equation 29 of [1] (page 8 of 21)

    if not calculate_via_put_call_parity and (cp == 1).any():
        Uk_call = calculate_Uk_european_options(cp=1, a=a, b=b, k=k)[:, np.newaxis] # Per equation 29 of [1] (page 8 of 21)

    if calculate_via_put_call_parity:
        # Method by put-call parity is more stable
        put_px = K * np.multiply(Fk, Uk_put).sum(axis=0) * np.exp(-r * tau)
        call_px = put_px + S0 * np.exp(-q * tau) - K * np.exp(-r * tau)
    else:
        if (cp == 1).any():
            call_px = K * np.multiply(Fk, Uk_call).sum(axis=0) * np.exp(-r * tau)
        if (cp == -1).any():
            put_px = K * np.multiply(Fk, Uk_put).sum(axis=0) * np.exp(-r * tau)

    result[cp == 1] = call_px[cp == 1]
    result[cp == -1] = put_px[cp == -1]

    return result


def heston_lipton_price_vanilla_european(S0, tau, r, q, cp, K, var0, vv, kappa, theta, rho):
    """
    European option price in the Heston model obtained using the Lewis-Lipton formula.

    Parameters:
        S0 (float): Initial asset price.
        tau (float): Time to maturity (in years).
        r (float): Risk-free interest rate (Domestic interest rate for Garman-Kohlhagen)
        q (float): Interest rate of the asset (Foreign interest rate for Garman-Kohlhagen)
        cp (int): 1 for a call option, -1 for a put option.
        K (float): Strike price.
        var0 (float): Initial variance.
        vv (float): Volatility of volatility.
        kappa (float): Mean reversion speed to the long-run variance.
        theta (float): Long-run variance.
        rho (float): Correlation between asset price and asset volatility Wiener processes.

    Returns:
        float: Option price.
    """

    integral_result, _ = scipy.integrate.quad(func=lambda v: heston_lipton_vanilla_european_integral(v=v, S0=S0, tau=tau, r=r, q=q, K=K, var0=var0, vv=vv, kappa=kappa, theta=theta, rho=rho), a=0, b=np.inf, epsrel=1e-8)
    C = np.exp(-q * tau) * S0 - np.exp(-r * tau) * K / np.pi * integral_result

    if cp == 1:
        P = C  # Call option price
    else:
        # Put option price via put-call parity
        P = C - S0 * np.exp(-q * tau) + K * np.exp(-r * tau)

    return P

@njit(fastmath=True, cache=True)
def heston_lipton_vanilla_european_integral(v, S0, tau, r, q, K, var0, vv, kappa, theta, rho):
    """
    Auxiliary function used by HestonVanillaLipton.

    Parameters:
        v (float): Integration variable.
        (Other parameters as in heston_lipton_price_fx_vanilla_european)

    Returns:
        float: Value of the integrand at point v.
    """
    X = np.log(S0 / K) + (r - q) * tau
    kappa_hat = kappa - rho * vv / 2
    zeta_term = v**2 * vv**2 * (1 - rho**2) + 2j * v * vv * rho * kappa_hat + kappa_hat**2 + vv**2 / 4
    zeta = np.sqrt(zeta_term)

    psi_plus = - (1j * kappa * rho * vv + kappa_hat) + zeta
    psi_minus = (1j * kappa * rho * vv + kappa_hat) + zeta

    numerator = psi_minus + psi_plus * np.exp(-zeta * tau)
    denominator = 2 * zeta
    log_term = np.log(numerator / denominator)

    alpha = - kappa * theta / vv**2 * (psi_plus * tau + 2 * log_term)
    beta = (1 - np.exp(-zeta * tau)) / (psi_minus + psi_plus * np.exp(-zeta * tau))

    exponent = (-1j * v + 0.5) * X + alpha - (v**2 + 0.25) * beta * var0
    payoff = np.real(np.exp(exponent)) / (v**2 + 0.25)

    return payoff


def heston_simulate_scalar(
        S0: float,
        mu: float,
        var0: float,
        vv: float,
        kappa: float,
        theta: float,
        rho: float,
        tau: float,
        rand_nbs: np.array,
        method: str = 'quadratic_exponential') -> np.ndarray:
    """
    Simulate scalar (non-vectorised) trajectories of the spot price and volatility processes using the Heston model.

    Parameters:
    ----------
    S0 : float
        Initial spot price.
    mu : float
        Drift term (r - q).
    var0 : float
        Initial variance.
    vv : float
        Volatility of volatility.
    kappa : float
        Speed of mean reversion for volatility.
    theta : float
        Long-run variance.
    rho : float
        Correlation between spot price and volatility.
    tau : float
        Time end point of the simulation.
    rand_nbs : np.ndarray
        Array of pseudorandom numbers for spot and volatility shocks. Shape=(# of timesteps, 2).
    method : str, optional
        The method used for simulating variance paths. Defaults to 'quadratic_exponential'.
        Options:
        - 'quadratic_exponential'
        - 'euler_with_absorption_of_volatility_process'
        - 'euler_with_reflection_of_volatility_process'

    Returns:
    -------
    np.ndarray
        Simulated trajectories of spot price and variance. Shape is (nb_timesteps + 1, 2).

    References:
    ----------
    [1] Janek, A., Kluge, T., Weron, R., Wystup, U. (2010). "FX smile in the Heston model".
    """

    assert rand_nbs.shape[1] == 2
    nb_timesteps = rand_nbs.shape[0]
    dt = tau / nb_timesteps
    x = np.zeros((nb_timesteps + 1, rand_nbs.shape[1]))

    C = np.array([[1, rho], [rho, 1]])
    u = normal_corr(C, rand_nbs) * np.sqrt(dt)

    if method == 'quadratic_exponential':
        x[0, :] = [np.log(S0), var0]
        phiC = 1.5

        for i in range(1, nb_timesteps + 1):
            m = theta + (x[i - 1, 1] - theta) * np.exp(-kappa * dt)
            s2 = (x[i - 1, 1] * vv ** 2 * np.exp(-kappa * dt) / kappa * (1 - np.exp(-kappa * dt)) +
                  theta * vv ** 2 / (2 * kappa) * (1 - np.exp(-kappa * dt)) ** 2)
            phi = s2 / m ** 2
            gamma1 = gamma2 = 0.5
            K0 = -rho * kappa * theta / vv * dt
            K1 = gamma1 * dt * (kappa * rho / vv - 0.5) - rho / vv
            K2 = gamma2 * dt * (kappa * rho / vv - 0.5) + rho / vv
            K3 = gamma1 * dt * (1 - rho ** 2)
            K4 = gamma2 * dt * (1 - rho ** 2)

            if phi <= phiC:
                b2 = 2 / phi - 1 + np.sqrt(2 / phi * (2 / phi - 1))
                a = m / (1 + b2)
                x[i, 1] = a * (np.sqrt(b2) + rand_nbs[i - 1, 1]) ** 2
            else:
                p = (phi - 1) / (phi + 1)
                beta = (1 - p) / m
                if 0 <= norm.cdf(rand_nbs[i - 1, 1]) <= p:
                    x[i, 1] = 0
                elif p < norm.cdf(rand_nbs[i - 1, 1]) <= 1:
                    x[i, 1] = 1 / beta * np.log((1 - p) / (1 - norm.cdf(rand_nbs[i - 1, 1])))

            x[i, 0] = x[i - 1, 0] + mu * dt + K0 + K1 * x[i - 1, 1] + K2 * x[i, 1] + \
                      np.sqrt(K3 * x[i - 1, 1] + K4 * x[i, 1]) * rand_nbs[i - 1, 0]

        x[:, 0] = np.exp(x[:, 0])
    elif method[:5] == 'euler':
        x[0, :] = [S0, var0]
        for i in range(1, nb_timesteps + 1):
            if method == 'euler_with_absorption_of_volatility_process':
                x[i - 1, 1] = max(x[i - 1, 1], 0)
            elif method == 'euler_with_reflection_of_volatility_process':
                x[i - 1, 1] = abs(x[i - 1, 1])

            x[i, 1] = x[i - 1, 1] + kappa * (theta - x[i - 1, 1]) * dt + vv * np.sqrt(x[i - 1, 1]) * u[i - 1, 1]
            x[i, 0] = x[i - 1, 0] + x[i - 1, 0] * (mu * dt + np.sqrt(x[i - 1, 1]) * u[i - 1, 0])

    return x


def heston_simulate(
        S0: float,
        mu: float,
        var0: float,
        vv: float,
        kappa: float,
        theta: float,
        rho: float,
        tau: float,
        rand_nbs: np.array,
        method: str = 'quadratic_exponential'):
    """
    Simulation trajectories of the spot price and volatility processes using the Heston model.
    Vectorised for multiple strikes.

    Parameters:
    ----------
    S0 : float
        Initial spot price.
    mu : float
        Drift term (r - q).
    var0 : float
        Initial variance.
    vv : float
        Volatility of volatility.
    kappa : float
        Speed of mean reversion for volatility.
    theta : float
        Long-run variance.
    rho : float
        Correlation between spot price and volatility.
    tau : float
        Time end point of the simulation.
    rand_nbs : np.ndarray
        Array of pseudorandom numbers for spot and volatility shocks. Shape is (# of timesteps, 2, # of simulations).
    method : str, optional
        The method used for simulating variance paths. Defaults to 'quadratic_exponential'.
        Options:
        - 'quadratic_exponential'
        - 'euler_with_absorption_of_volatility_process'
        - 'euler_with_reflection_of_volatility_process'

    Returns:
    -------
    np.ndarray
        Simulated trajectories of spot price and variance. Shape is (nb_timesteps + 1, 2).

    References:
    ----------
    [1] Janek, A., Kluge, T., Weron, R., Wystup, U. (2010). "FX smile in the Heston model".
     """

    assert rand_nbs.shape[1] == 2
    nb_timesteps = rand_nbs.shape[0]
    dt = tau / nb_timesteps
    x = np.zeros((nb_timesteps + 1, rand_nbs.shape[1], rand_nbs.shape[2]))

    C = np.array([[1, rho], [rho, 1]])
    u = normal_corr(C, rand_nbs) * np.sqrt(dt)  # need to check this correlation works as expected

    if method == 'quadratic_exponential':
        x[0, 0, :] = np.log(S0)
        x[0, 1, :] = var0

        phiC = 1.5

        for i in range(1, nb_timesteps + 1):
            m = theta + (x[i - 1, 1, :] - theta) * np.exp(-kappa * dt)
            s2 = (x[i - 1, 1, :] * vv ** 2 * np.exp(-kappa * dt) / kappa * (1 - np.exp(-kappa * dt)) +
                  theta * vv ** 2 / (2 * kappa) * (1 - np.exp(-kappa * dt)) ** 2)
            phi = s2 / m ** 2
            gamma1 = gamma2 = 0.5
            K0 = -rho * kappa * theta / vv * dt
            K1 = gamma1 * dt * (kappa * rho / vv - 0.5) - rho / vv
            K2 = gamma2 * dt * (kappa * rho / vv - 0.5) + rho / vv
            K3 = gamma1 * dt * (1 - rho ** 2)
            K4 = gamma2 * dt * (1 - rho ** 2)

            mask = phi <= phiC
            b2 = np.full(mask.shape, np.nan)
            b2[mask] = 2 / phi[mask] - 1 + np.sqrt(2 / phi[mask] * (2 / phi[mask] - 1))
            a = np.full(mask.shape, np.nan)
            a[mask] = m[mask] / (1 + b2[mask])
            x[i, 1, mask] = a[mask] * (np.sqrt(b2[mask]) + rand_nbs[i - 1, 1, mask]) ** 2

            mask = np.logical_not(mask)
            p = np.full(mask.shape, np.nan)
            p[mask] = (phi[mask] - 1) / (phi[mask] + 1)
            beta = np.full(mask.shape, np.nan)
            beta[mask] = (1 - p[mask]) / m[mask]

            mask_0_p = np.logical_and(0 <= norm.cdf(rand_nbs[i - 1, 1, :]), norm.cdf(rand_nbs[i - 1, 1, :]) <= p)
            mask_net = np.logical_and(mask, mask_0_p)
            x[i, 1, mask_net] = 0

            mask_p_1 = np.logical_and(p < norm.cdf(rand_nbs[i - 1, 1, :]), norm.cdf(rand_nbs[i - 1, 1, :]) <= 1)
            mask_net = np.logical_and(mask, mask_p_1)
            x[i, 1, mask_net] = 1 / beta[mask_net] * np.log(
                (1 - p[mask_net]) / (1 - norm.cdf(rand_nbs[i - 1, 1, mask_net])))

            x[i, 0, :] = x[i - 1, 0, :] + mu * dt + K0 + K1 * x[i - 1, 1, :] + K2 * x[i, 1, :] + \
                         np.sqrt(K3 * x[i - 1, 1, :] + K4 * x[i, 1, :]) * rand_nbs[i - 1, 0, :]

        x[:, 0] = np.exp(x[:, 0])

    elif method[:5] == 'euler':
        x[0, 0, :] = S0
        x[0, 1, :] = var0

        for i in range(1, nb_timesteps + 1):
            if method == 'euler_with_absorption_of_volatility_process':
                x[i - 1, 1, :] = np.maximum(0, x[i - 1, 1, :])
            elif method == 'euler_with_reflection_of_volatility_process':
                x[i - 1, 1, :] = np.abs(x[i - 1, 1, :])

            x[i, 1, :] = x[i - 1, 1, :] \
                         + kappa * (theta - x[i - 1, 1, :]) * dt + vv * np.sqrt(x[i - 1, 1, :]) * u[i - 1, 1, :]
            x[i, 0, :] = x[i - 1, 0, :] \
                         + x[i - 1, 0, :] * (mu * dt + np.sqrt(x[i - 1, 1, :]) * u[i - 1, 0, :])

    return x




























# -*- coding: utf-8 -*-
import os

from frm.utils import workday

if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM')) 

from frm.enums import FXSmileInterpolationMethod, DayCountBasis, CompoundingFreq
from frm.term_structures.zero_curve import ZeroCurve
from frm.pricing_engine.monte_carlo_generic import generate_rand_nbs
from frm.pricing_engine.heston import heston_calibrate_vanilla_smile, heston_price_vanilla_european, heston_simulate
from frm.pricing_engine.garman_kohlhagen import garman_kohlhagen_price, garman_kohlhagen_solve_strike_from_delta, garman_kohlhagen_solve_implied_vol
from frm.pricing_engine.geometric_brownian_motion import simulate_gbm_path
from frm.pricing_engine.volatility_generic import forward_volatility, flat_forward_interp
from frm.term_structures.fx_helpers import (clean_vol_quotes_column_names,
                                            get_delta_smile_quote_details,
                                            check_delta_convention,
                                            fx_term_structure_helper,
                                            fx_forward_curve_helper,
                                            interp_fx_forward_curve_df,
                                            validate_ccy_pair,
                                            solve_call_put_quotes_from_strategy_quotes)

from scipy.interpolate import CubicSpline, InterpolatedUnivariateSpline
from frm.utils import year_frac, get_busdaycal, resolve_fx_curve_dates

import numpy as np
import pandas as pd
from dataclasses import dataclass, field, InitVar
from typing import Optional
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick


@dataclass
class FXVolatilitySurface:
    # Mandatory initialisation attributes
    domestic_ccy: str
    foreign_ccy: str
    fx_forward_curve_df: pd.DataFrame # Quotes in perspective of # of domestic currency units per foreign currency unit
    domestic_zero_curve: ZeroCurve
    foreign_zero_curve: ZeroCurve
    vol_quotes: InitVar[pd.DataFrame]

    #### Optional initialisation attributes

    # One of curve_date or spot_date must be provided. If spot_date is not provided, it will be calculated from curve_date and spot_offset or set per market convention
    curve_date: Optional[pd.Timestamp] = None
    spot_offset: Optional[int] = None
    spot_date: Optional[pd.Timestamp] = None

    cal: np.busdaycalendar = None
    day_count_basis: DayCountBasis = DayCountBasis.ACT_ACT
    smile_interpolation_method: FXSmileInterpolationMethod = FXSmileInterpolationMethod.UNIVARIATE_SPLINE

    # Non initialisation attributes set in __post_init__
    fx_spot_rate: float = field(init=False)
    vol_smile_pillar_df: pd.DataFrame = field(init=False)
    vol_smile_daily_df: pd.DataFrame = field(init=False)
    vol_smile_daily_func: dict = field(init=False)

    quotes_column_names: np.ndarray = field(init=False)
    quotes_signed_delta: np.ndarray = field(init=False)
    quotes_call_put_flag: np.ndarray = field(init=False)


    def __post_init__(self, vol_quotes):

        self.vol_smile_daily_func = {}

        self.ccy_pair, self.domestic_ccy, self.foreign_ccy =  validate_ccy_pair(self.foreign_ccy + self.domestic_ccy)

        if self.cal is None:
            self.cal = get_busdaycal([self.domestic_ccy, self.foreign_ccy])

        self.curve_date, self.spot_offset, self.spot_date = resolve_fx_curve_dates(self.ccy_pair,
                                                                                   self.cal,
                                                                                   self.curve_date,
                                                                                   self.spot_offset,
                                                                                   self.spot_date)
        assert self.curve_date == self.domestic_zero_curve.curve_date
        assert self.curve_date == self.foreign_zero_curve.curve_date

        self.fx_forward_curve_df = fx_forward_curve_helper(fx_forward_curve_df=self.fx_forward_curve_df,
                                                           curve_date=self.curve_date,
                                                           spot_offset=self.spot_offset,
                                                           cal=self.cal)
        mask = self.fx_forward_curve_df['tenor'] == 'sp'
        self.fx_spot_rate = self.fx_forward_curve_df.loc[mask, 'fx_forward_rate'].values[0]

        # Process volatility input and setup pillar dataframe and daily helper dataframe
        vol_quotes = clean_vol_quotes_column_names(vol_quotes)
        vol_quotes_call_put = solve_call_put_quotes_from_strategy_quotes(vol_quotes)
        delta_smile_quote_details = get_delta_smile_quote_details(vol_quotes_call_put)
        self.quotes_column_names = np.array(delta_smile_quote_details['quotes_column_names'].to_list())
        self.quotes_signed_delta = np.array(delta_smile_quote_details['quotes_signed_delta'].to_list())
        self.quotes_call_put_flag = np.array(delta_smile_quote_details['quotes_call_put_flag'].to_list())
        self.vol_smile_pillar_df = self._setup_vol_smile_pillar_df(vol_quotes_call_put)
        self.vol_smile_daily_df = self._setup_vol_smile_daily_df()
        self.strike_pillar_df = self._setup_strike_pillar()


    def _add_fx_ir_to_df(self, df: pd.DataFrame) -> pd.DataFrame:
        df['fx_forward_rate'] = interp_fx_forward_curve_df(fx_forward_curve_df=self.fx_forward_curve_df,
                                                           dates=df['expiry_date'],
                                                           date_type='fixing_date')
        df['domestic_zero_rate'] = self.domestic_zero_curve.get_zero_rates(dates=df['expiry_date'],
                                                                          compounding_freq=CompoundingFreq.CONTINUOUS)
        df['foreign_zero_rate'] = self.foreign_zero_curve.get_zero_rates(dates=df['expiry_date'],
                                                                        compounding_freq=CompoundingFreq.CONTINUOUS)
        return df


    def _setup_vol_smile_pillar_df(self, vol_smile_pillar_df):
            vol_smile_pillar_df['warnings'] = ''

            vol_smile_pillar_df = check_delta_convention(vol_smile_pillar_df, self.ccy_pair)
            vol_smile_pillar_df = fx_term_structure_helper(df=vol_smile_pillar_df,
                                                           curve_date=self.curve_date,
                                                           spot_offset=self.spot_offset,
                                                           cal=self.cal,
                                                           rate_set_date_str='expiry_date')
            vol_smile_pillar_df['expiry_years'] = year_frac(self.curve_date, vol_smile_pillar_df['expiry_date'], self.day_count_basis)
            vol_smile_pillar_df = self._add_fx_ir_to_df(vol_smile_pillar_df)
            column_order = ['warnings'] \
                           + (['tenor'] if 'tenor' in vol_smile_pillar_df.columns else []) \
                           + ['expiry_date', 'delivery_date', 'expiry_years', 'fx_forward_rate', 'domestic_zero_rate',
                              'foreign_zero_rate', 'delta_convention'] \
                           + list(self.quotes_column_names)
            return vol_smile_pillar_df[column_order]


    def _setup_vol_smile_daily_df(self):
        min_expiry = self.vol_smile_pillar_df['expiry_date'].min()
        max_expiry = self.vol_smile_pillar_df['expiry_date'].max()
        expiry_dates = pd.date_range(min_expiry, max_expiry, freq='d')
        delivery_dates = workday(expiry_dates, offset=self.spot_offset, cal=self.cal)

        expiry_years = year_frac(self.curve_date, expiry_dates, self.day_count_basis)

        vol_smile_daily_df = pd.DataFrame({'expiry_date_daily': expiry_dates,
                                           'delivery_date_daily': delivery_dates,
                                           'expiry_years_daily': expiry_years})

        vol_smile_pillar_df = self.vol_smile_pillar_df.copy()
        vol_smile_pillar_df['expiry_years'] = year_frac(self.curve_date, vol_smile_pillar_df['expiry_date'],self.day_count_basis)

        # Merge to find closest smaller and larger tenors for each target tenor
        lower_pillar = pd.merge_asof(vol_smile_daily_df, vol_smile_pillar_df, left_on='expiry_date_daily',
                                     right_on='expiry_date', direction='backward')
        upper_pillar = pd.merge_asof(vol_smile_daily_df, vol_smile_pillar_df, left_on='expiry_date_daily',
                                     right_on='expiry_date', direction='forward')

        # TODO: Adjust the interpolation to consider the delta convention where pillar points have different delta conventions
        vol_smile_daily_df['delta_convention'] = upper_pillar['delta_convention']

        # Convert to numpy for efficient calculations
        t1 = lower_pillar['expiry_years'].to_numpy()
        t2 = upper_pillar['expiry_years'].to_numpy()
        t = vol_smile_daily_df['expiry_years_daily'].to_numpy()
        t1 = t1[:, np.newaxis]
        t2 = t2[:, np.newaxis]
        t = t[:, np.newaxis]
        vol_t1 = lower_pillar[self.quotes_column_names].to_numpy()
        vol_t2 = upper_pillar[self.quotes_column_names].to_numpy()

        vol_smile_daily_df.rename(columns={'expiry_date_daily': 'expiry_date',
                                           'expiry_years_daily': 'expiry_years'}, inplace=True)

        # Interpolate the volatility smile
        vol_smile_daily_df[self.quotes_column_names] = flat_forward_interp(t1, vol_t1, t2, vol_t2, t)

        vol_smile_daily_df = self._add_fx_ir_to_df(vol_smile_daily_df)

        column_order = ['expiry_date', 'expiry_years', 'fx_forward_rate', 'domestic_zero_rate', 'foreign_zero_rate', 'delta_convention'] + list(self.quotes_column_names)

        return vol_smile_daily_df[column_order]


    def _setup_strike_pillar(self):
        strike_pillar_df = self.vol_smile_pillar_df.copy()
        strike_pillar_df.loc[:, self.quotes_column_names] = np.nan
        for i, row in strike_pillar_df.iterrows():
            strikes = garman_kohlhagen_solve_strike_from_delta(
                S0=self.fx_spot_rate,
                tau=row['expiry_years'],
                r_d=row['domestic_zero_rate'],
                r_f=row['foreign_zero_rate'],
                vol=self.vol_smile_pillar_df.loc[i, self.quotes_column_names].values,
                signed_delta=self.quotes_signed_delta,
                delta_convention=row['delta_convention'],
                F=row['fx_forward_rate'])
            strike_pillar_df.loc[i, self.quotes_column_names] = strikes
        return strike_pillar_df


    def _solve_vol_daily_smile_func(self, expiry_dates: pd.DatetimeIndex):
        """
        Internal method for solving the function definition for the volatility smile.
        The 'function' is a spline or the Heston parameters that allow for the interpolation of the volatility smile for any strike/delta.
        The function is solved for each date in `expiry_dates` and stored in the `vol_smile_daily_func` attribute.

        Parameters:
        dates (pd.DatetimeIndex): Array of dates to perform the σ surface interpolation for.

        Raises:
        ValueError: If the sum of squared errors (SSE) from the Heston fit exceeds a threshold, indicating a poor fit.
        """

        for expiry_date in expiry_dates.unique():
            if expiry_date not in self.vol_smile_daily_func.keys():

                mask = self.vol_smile_daily_df['expiry_date'] == expiry_date
                row = self.vol_smile_daily_df.loc[mask].copy()

                # Scalars
                S0 = self.fx_spot_rate
                r_f=row['foreign_zero_rate'].iloc[0]
                r_d=row['domestic_zero_rate'].iloc[0]
                tau=row['expiry_years'].iloc[0]
                delta_convention=row['delta_convention'].iloc[0]
                F=row['fx_forward_rate'].iloc[0]

                if F is not None:
                    # Use market forward rate and imply the curry basis-adjusted domestic interest rate
                    F = np.atleast_1d(F).astype(float)
                    r_d_basis_adj = np.log(F / S0) / tau + r_f  # from F = S0 * exp((r_d - r_f) * tau)
                    r = r_d_basis_adj
                    q = r_f
                else:
                    r = r_d
                    q = r_f

                # Arrays
                vol = row[self.quotes_column_names].iloc[0].values
                signed_delta = self.quotes_signed_delta
                cp = self.quotes_call_put_flag

                if self.smile_interpolation_method in [FXSmileInterpolationMethod.UNIVARIATE_SPLINE,
                                                       FXSmileInterpolationMethod.CUBIC_SPLINE]:
                    K = garman_kohlhagen_solve_strike_from_delta(
                        S0=S0, tau=tau, r_d=r_d, r_f=r_f, vol=vol, signed_delta=signed_delta, delta_convention=delta_convention, F=F)
                    if self.smile_interpolation_method == FXSmileInterpolationMethod.UNIVARIATE_SPLINE:
                        if len(self.quotes_column_names) < 3:
                            raise ValueError('Cannot fit InterpolatedUnivariateSpline with less than 3 points, please provide more points.')
                        degree = min(3, max(2, len(self.quotes_column_names) - 1))
                        self.vol_smile_daily_func[expiry_date] = InterpolatedUnivariateSpline(x=K, y=vol, k=degree)
                    elif self.smile_interpolation_method == FXSmileInterpolationMethod.CUBIC_SPLINE:
                        if len(self.quotes_column_names) < 4:
                            raise ValueError('Cannot fit CubicSpline with less than 4 points, please provide more points or use a different interpolation method')
                        self.vol_smile_daily_func[expiry_date] = CubicSpline(x=K, y=vol)
                elif self.smile_interpolation_method in [FXSmileInterpolationMethod.HESTON_1993,
                                                         FXSmileInterpolationMethod.HESTON_CARR_MADAN_GAUSS_KRONROD_QUADRATURE,
                                                         FXSmileInterpolationMethod.HESTON_CARR_MADAN_FFT_W_SIMPSONS,
                                                         FXSmileInterpolationMethod.HESTON_COSINE,
                                                         FXSmileInterpolationMethod.HESTON_LIPTON]:
                    var0, vv, kappa, theta, rho, lambda_, IV, SSE = \
                        heston_calibrate_vanilla_smile(volatility_quotes=vol,
                                                       delta_of_quotes=signed_delta,
                                                       S0=S0,
                                                       r=r,
                                                       q=q,
                                                       tau=tau,
                                                       cp=cp,
                                                       delta_convention=delta_convention,
                                                       pricing_method=self.smile_interpolation_method.value)

                    if SSE < 0.001:
                        result = {
                            'var0': var0,
                            'vv': vv,
                            'kappa': kappa,
                            'theta': theta,
                            'rho': rho,
                            'lambda_': lambda_,
                            'IV': IV,
                            'SSE': SSE
                        }
                        self.vol_smile_daily_func[expiry_date] = result
                    else:
                        raise ValueError('SSE is a large value, ', round(SSE, 4), ' heston fit at ', expiry_date,
                                         ' is likely poor')


    def interp_vol_surface(self,
                         expiry_dates: pd.DatetimeIndex,
                         K: np.ndarray,
                         cp: np.ndarray) -> pd.DataFrame:

        K = np.atleast_1d(K).astype(float)
        cp = np.atleast_1d(cp).astype(float)
        assert expiry_dates.shape == K.shape
        assert expiry_dates.shape == cp.shape

        self._solve_vol_daily_smile_func(expiry_dates)

        mask = self.vol_smile_daily_df['expiry_date'].isin(expiry_dates)
        tmp_df = self.vol_smile_daily_df.loc[mask].copy()
        tmp_df.set_index('expiry_date', inplace=True)
        interp_df = tmp_df.loc[expiry_dates].copy()
        interp_df.index.name = 'expiry_date'
        interp_df.reset_index(inplace=True, drop=False)
        interp_df['K'] = K
        interp_df['cp'] = cp
        interp_df['vol'] = np.nan

        for i,row in interp_df.iterrows():
            vol_smile_func = self.vol_smile_daily_func[row['expiry_date']]

            if self.smile_interpolation_method in [FXSmileInterpolationMethod.UNIVARIATE_SPLINE,
                                                    FXSmileInterpolationMethod.CUBIC_SPLINE]:
                interp_df.loc[i,'vol'] = vol_smile_func(K[i])
            elif self.smile_interpolation_method in [FXSmileInterpolationMethod.HESTON_1993,
                                                    FXSmileInterpolationMethod.HESTON_CARR_MADAN_GAUSS_KRONROD_QUADRATURE,
                                                    FXSmileInterpolationMethod.HESTON_CARR_MADAN_FFT_W_SIMPSONS,
                                                    FXSmileInterpolationMethod.HESTON_COSINE,
                                                    FXSmileInterpolationMethod.HESTON_LIPTON]:

                S0 = self.fx_spot_rate
                r_f = row['foreign_zero_rate']
                r_d = row['domestic_zero_rate']
                tau = row['expiry_years']
                F = row['fx_forward_rate']

                if F is not None:
                    # Use market forward rate and imply the curry basis-adjusted domestic interest rate
                    F = np.atleast_1d(F).astype(float)
                    r_d_basis_adj = np.log(F / S0) / tau + r_f  # from F = S0 * exp((r_d - r_f) * tau)
                    r = r_d_basis_adj
                    q = r_f
                else:
                    r = r_d
                    q = r_f

                X = heston_price_vanilla_european(
                    S0=S0, tau=tau, r=r, q=q, cp=cp[i], K=K[i], var0=vol_smile_func['var0'],
                    vv=vol_smile_func['vv'], kappa=vol_smile_func['kappa'], theta=vol_smile_func['theta'],
                    rho=vol_smile_func['rho'], lambda_=vol_smile_func['lambda_'],
                    pricing_method=self.smile_interpolation_method.value)

                interp_df.loc[i,'vol'] = garman_kohlhagen_solve_implied_vol(
                    S0=S0, tau=tau, r_d=r, r_f=q, cp=cp[i], K=K[i], X=X,
                    vol_guess=np.sqrt(vol_smile_func['var0']))

        return interp_df


    def price_vanilla_european(self,
                               expiry_dates: pd.DatetimeIndex,
                               K: [float, np.ndarray],
                               cp: [float, np.ndarray],
                               analytical_greeks_flag: bool=False,
                               numerical_greeks_flag: bool=False,
                               intrinsic_time_split_flag: bool=False) -> dict:

        K, cp = map(np.atleast_1d, (K, cp))
        assert expiry_dates.shape == K.shape
        assert expiry_dates.shape == cp.shape

        interp_df = self.interp_vol_surface(expiry_dates=expiry_dates, K=K, cp=cp)

        results = garman_kohlhagen_price(
            S0=self.fx_spot_rate,
            tau=interp_df['expiry_years'].values,
            r_d=interp_df['domestic_zero_rate'].values,
            r_f=interp_df['foreign_zero_rate'].values,
            cp=cp,
            K=K,
            vol=interp_df['vol'].values,
            F=interp_df['fx_forward_rate'].values,
            analytical_greeks=analytical_greeks_flag,
            numerical_greeks=numerical_greeks_flag,
            intrinsic_time_split=intrinsic_time_split_flag
        )

        results['market_data_inputs'] = pd.DataFrame({'vol': interp_df['vol'].values,
                                                      'r_d': interp_df['domestic_zero_rate'].values,
                                                      'r_f': interp_df['foreign_zero_rate'].values,
                                                      'F': interp_df['fx_forward_rate'].values})

        return results


    def simulate_gbm_fx_rate_path(self,
                                  delivery_date_grid: pd.DatetimeIndex,
                                  nb_simulations: Optional[int]=None,
                                  apply_antithetic_variates: bool=True):

        delivery_date_grid = delivery_date_grid.unique().sort_values(ascending=True)
        delivery_date_grid = delivery_date_grid.union(self.spot_date)
        fixing_date_grid = np.busday_offset(delivery_date_grid, offsets=-1*self.spot_offset, roll='preceding', busdaycal=self.cal)
        mask = self.vol_smile_daily_df['expiry_date'].isin(fixing_date_grid)
        vol_smile_daily_df = self.vol_smile_daily_df.loc[mask].copy()

        schedule = pd.DataFrame({
            'fixing_date': fixing_date_grid,
            'delivery_date': delivery_date_grid,
            'fixing_years': year_frac(self.curve_date, delivery_date_grid, self.day_count_basis).values,
            'domestic_zero_rates': vol_smile_daily_df['domestic_zero_rate'].values,
            'foreign_zero_rates': vol_smile_daily_df['foreign_zero_rate'].values,
            'fx_forward_rates:': vol_smile_daily_df['fx_forward_rate'].values,
        })

        if 'atm_forward' in self.vol_smile_daily_df.columns:
            schedule['atm_forward'] = vol_smile_daily_df['atm_forward'].values
        elif 'atm_delta_neutral' in self.vol_smile_daily_df.columns:
            schedule['volatility'] = vol_smile_daily_df['atm_delta_neutral'].values
        else:
            raise ValueError

        schedule['dt'] = schedule['fixing_years'].diff()
        schedule['dt'].iloc[0] = 0.0

        schedule['drift'] = np.log(schedule['fx_forward_rates:'].shift(-1) / schedule['fx_forward_rates:']) / schedule['dt']
        schedule['drift'].iloc[0] = np.nan

        schedule['forward_volatility'] = forward_volatility(
            t1=schedule['fixing_years'].values[:-1],
            vol_t1=schedule['volatility'].values[:-1],
            t2=schedule['fixing_years'].values[1:],
            vol_t2=schedule['volatility'].values[1:]
        )

        results = {'gbm_monte_carlo_market_data_inputs': schedule}

        rand_nbs = generate_rand_nbs(nb_steps=len(delivery_date_grid) - 1,
                                     nb_rand_vars=1,
                                     nb_simulations=nb_simulations,
                                     apply_antithetic_variates=apply_antithetic_variates)

        fx_rate_simulation_paths = simulate_gbm_path(initial_px=self.fx_spot_rate,
                                                     drift=schedule['drift'].values[1:],
                                                     forward_volatility=schedule['forward_volatility'].values[1:],
                                                     timestep_length=schedule['dt'].values[1:],
                                                     rand_nbs=rand_nbs)

        results['fx_rate_simulation_paths'] = fx_rate_simulation_paths

        return results


    def simulate_heston_fx_rate_path(self,
                                     delivery_date_grid,
                                     nb_simulations=None,
                                     flag_apply_antithetic_variates=True):


        delivery_date_grid = delivery_date_grid.unique().sort_values(ascending=True)
        delivery_date_grid = delivery_date_grid.union(self.spot_date)
        fixing_date_grid = np.busday_offset(delivery_date_grid, offsets=-1*self.spot_offset, roll='preceding', busdaycal=self.cal)
        #mask = self.vol_smile_daily_df['expiry_date'].isin(fixing_date_grid)
        #vol_smile_daily_df = self.vol_smile_daily_df.loc[mask].copy()
        self._solve_vol_daily_smile_func(fixing_date_grid)

        # It doesn't really make sense to simulate a rate path using the Heston model.
        # The heston model is designed to for fit the volatility smile at a certain point in time, not the term structure of volatility along the path.

        results = {}

        for i, (fixing_date, delivery_date) in enumerate(zip(fixing_date_grid,delivery_date_grid)):
            print(i, fixing_date, delivery_date)

            expiry_years = year_frac(self.curve_date, fixing_date, self.day_count_basis)
            fx_forward_rate = interp_fx_forward_curve_df(self.fx_forward_curve_df,
                                                         dates=fixing_date,
                                                         date_type='fixing_date').values

            mu = np.log(fx_forward_rate / self.fx_spot_rate) / expiry_years

            var0 = self.vol_smile_daily_func[fixing_date]['var0']
            vv = self.vol_smile_daily_func[fixing_date]['vv']
            kappa = self.vol_smile_daily_func[fixing_date]['kappa']
            theta = self.vol_smile_daily_func[fixing_date]['theta']
            rho = self.vol_smile_daily_func[fixing_date]['rho']
            # lambda_ is not used in the simulation, used in the calibration only.
            # lambda_ = self.daily_volatility_smile_func[fixing_date]['lambda_']

            rand_nbs = generate_rand_nbs(
                nb_steps=100,
                nb_rand_vars=2,
                nb_simulations=10 * 1000,
                apply_antithetic_variates=False)

            sim_results = heston_simulate(
                S0=self.fx_spot_rate,
                mu=mu,
                var0=var0,
                vv=vv,
                kappa=kappa,
                theta=theta,
                rho=rho,
                tau=expiry_years,
                rand_nbs=rand_nbs)

            results[delivery_date] = sim_results

        return results


    def plot_smile(self, expiry_date: pd.Timestamp):

        self._solve_vol_daily_smile_func(pd.DatetimeIndex([expiry_date]))

        vol_daily_smile_df = self.vol_smile_daily_df.loc[self.vol_smile_daily_df['expiry_date'] == expiry_date]

        delta_convention = vol_daily_smile_df['delta_convention'].iloc[0]
        vol_pillar = vol_daily_smile_df[self.quotes_column_names].iloc[0]

        pillar_strikes = garman_kohlhagen_solve_strike_from_delta(S0=self.fx_spot_rate,
                                  tau=vol_daily_smile_df['expiry_years'].iloc[0],
                                  r_d=vol_daily_smile_df['domestic_zero_rate'].iloc[0],
                                  r_f=vol_daily_smile_df['foreign_zero_rate'].iloc[0],
                                  vol=vol_pillar,
                                  signed_delta=self.quotes_signed_delta,
                                  delta_convention=delta_convention,
                                  F=vol_daily_smile_df['fx_forward_rate'].iloc[0],
                                  atm_delta_convention='forward')

        nb_strikes = 100
        strike_delta_neutral = pillar_strikes[self.quotes_column_names == 'atm_delta_neutral'].item()
        strikes_put = np.linspace(np.min(pillar_strikes), strike_delta_neutral, nb_strikes)
        strikes_call = np.linspace(strike_delta_neutral, np.max(pillar_strikes), nb_strikes)
        strikes_array = np.concatenate([strikes_put, strikes_call])
        call_put_array = np.concatenate([-1 * np.ones(nb_strikes), np.ones(nb_strikes)])
        expiry_dates = pd.DatetimeIndex([expiry_date for _ in range(len(strikes_array))])

        result = self.price_vanilla_european(expiry_dates=expiry_dates,
                                             K=strikes_array,
                                             cp=call_put_array,
                                             analytical_greeks_flag=True)

        if delta_convention == 'regular_spot':
            delta_str = 'spot_delta'
        elif delta_convention == 'regular_forward':
            delta_str = 'forward_delta'

        delta = result['analytical_greeks'][delta_str]
        vol = result['market_data_inputs']['vol']

        delta_pillar_for_plot = np.full(self.quotes_signed_delta.shape, np.nan)
        midpoint = len(delta_pillar_for_plot) // 2
        delta_pillar_for_plot[:midpoint] = np.abs(self.quotes_signed_delta[:midpoint])
        delta_pillar_for_plot[midpoint:] = 1.0 - self.quotes_signed_delta[midpoint:]

        delta_for_plot = np.full(delta.shape, np.nan)
        delta_for_plot[:nb_strikes] = np.abs(delta[:nb_strikes])
        delta_for_plot[nb_strikes:] = 1.0 - delta[nb_strikes:]

        # Plot
        plt.plot(100 * delta_pillar_for_plot, vol_pillar, 'x', label='Pillar Volatilities')
        plt.plot(100 * delta_for_plot, vol, label='Fitted Smile')  # X markers

        tenor = np.round(vol_daily_smile_df['expiry_years'].iloc[0], 2)

        plt.title(f'Volatility Smile for {tenor}Y ({expiry_date.date()})\n'
                  f'Smile interpolation method: {self.smile_interpolation_method.value}')
        plt.xlabel('Delta:   Puts: (0,0.5). ATM Δ Neutral: 0.5. Calls: (0.5,1)')
        plt.ylabel('Volatility')
        plt.xlim([0, 100])

        plt.gca().yaxis.set_major_formatter(mtick.PercentFormatter(1.0, decimals=1))  # 1.0 means 100%

        plt.legend()
        plt.show()



# -*- coding: utf-8 -*-
import os
if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM')) 

from scipy.interpolate import splrep, splev
import pandas as pd
import numpy as np
from dataclasses import dataclass, field, InitVar
from typing import Optional, Union
import matplotlib.pyplot as plt
import datetime as dt
from dateutil.relativedelta import relativedelta
# Own project imports
from frm.utils import day_count, year_frac, clean_tenor, tenor_to_date_offset, convert_column_to_consistent_data_type
from frm.enums import DayCountBasis, CompoundingFreq, RFRFixingCalcMethod, TermRate, ZeroCurveInterpMethod, ZeroCurveExtrapMethod
from frm.term_structures.zero_curve_helpers import zero_rate_from_discount_factor, discount_factor_from_zero_rate









@dataclass
class ZeroCurve: 
    # Required inputs
    curve_date: pd.Timestamp
    pillar_df: pd.DataFrame
    
    # Used in the __post_init__ but not set as attributes
    compounding_freq: InitVar[CompoundingFreq]=None # defines the zero_rate compounding frequency in 'data'
    
    # Optional init inputs
    day_count_basis: DayCountBasis=DayCountBasis.ACT_ACT
    cal: np.busdaycalendar=np.busdaycalendar() # used in simple-average forward rate calculation
    interp_method: ZeroCurveInterpMethod=ZeroCurveInterpMethod.LINEAR_ON_LN_DISCOUNT
    extrap_method: ZeroCurveExtrapMethod=ZeroCurveExtrapMethod.NONE


    # Attributes set in __post_init__
    cubic_spline_definition: tuple=field(init=False)
    
    def __post_init__(self, compounding_freq):
        
        pillar_df = self.pillar_df
        if compounding_freq is None and 'zero_rate' in pillar_df.columns:
            raise ValueError("'compounding_freq' must be specified zero rates specified in 'pillar_df'")
            
        self._process_pillar_df(pillar_df, compounding_freq)
        self._setup_daily_df()
    
    def _process_pillar_df(self, pillar_df, compounding_freq):

        only_one_of_columns_X = ['tenor', 'date', 'days', 'years']
        only_one_of_columns_Y = ['zero_rate','discount_factor']
        
        if len(pillar_df.columns.to_list()) != 2:
            raise ValueError('Exactly two columns must be specified: \n'
                             '(i) One of: ' + ', '.join(only_one_of_columns_X) + '\n' 
                             '(ii) One of: ' + ', '.join(only_one_of_columns_Y))            
        
        X_columns = [col for col in only_one_of_columns_X if col in pillar_df.columns]
        if len(X_columns) != 1:
            raise ValueError('Exactly one of the following columns must be specified: ' + ', '.join(only_one_of_columns_X))       
        X_column_name = X_columns[0]
        
        Y_columns  = [col for col in only_one_of_columns_Y if col in pillar_df.columns]
        if len(Y_columns) != 1:
            raise ValueError('Exactly one of the following columns must be specified: ' + ', '.join(only_one_of_columns_Y))
        Y_column_name = Y_columns[0]

        def __calculate_days(): 
            pillar_df['days'] = (pillar_df['date'] - self.curve_date).dt.days
        def __calculate_years(): 
            pillar_df['years'] = year_frac(self.curve_date, pillar_df['date'], self.day_count_basis)
        
        match X_column_name:
            case 'tenor':
                pillar_df.loc[:,'tenor'] = pillar_df['tenor'].apply(clean_tenor)
                date_offset = pillar_df['tenor'].apply(tenor_to_date_offset)
                dates = self.curve_date + date_offset
                pillar_df['date'] = np.busday_offset(dates.values.astype('datetime64[D]'), offsets=0, roll='following', busdaycal=self.cal)
                __calculate_days()
                __calculate_years()
            case 'date':
                __calculate_days()
                __calculate_years()
            case 'days':
                pillar_df['date'] =  pillar_df['days'].apply(lambda days: self.curve_date + dt.timedelta(days=days))
                __calculate_years()
            case 'years':
                pass
            
        pillar_df = pillar_df.sort_values(by='years', ascending=True).reset_index(drop=True)
        pillar_df = convert_column_to_consistent_data_type(pillar_df)
            
        if Y_column_name == 'zero_rate':
            pillar_df['discount_factor'] = discount_factor_from_zero_rate(
                years=pillar_df['years'],
                zero_rate=pillar_df['zero_rate'],
                compounding_freq=compounding_freq)
            pillar_df.drop(columns=['zero_rate'], inplace=True)

        # Add continuously compounded interest rate for internal use
        pillar_df['cczr'] = -1 * np.log(pillar_df['discount_factor']) / pillar_df['years']

        # Add pillar points at t=0.
        # This is a dependency to the HW1F class which requires a short rate at t=0
        first_row = {'years': 0.0, 'discount_factor': 1.0}
        if 'tenor' in pillar_df.columns:
            first_row['tenor'] = ''
        if 'date' in pillar_df.columns:
            first_row['date'] = self.curve_date
        if 'days' in pillar_df.columns:
            first_row['days'] = 0.0

        match self.interp_method:
            case ZeroCurveInterpMethod.LINEAR_ON_LN_DISCOUNT:
                first_row['cczr'] = pillar_df['cczr'].iloc[0]
                self.cubic_spline_definition = None
            case ZeroCurveInterpMethod.CUBIC_SPLINE_ON_CCZR:
                x = [0.0] + pillar_df['years'].to_list()
                y = [0.0] + np.log(pillar_df['discount_factor']).to_list()
                self.cubic_spline_definition = splrep(x=x, y=y, k=3)
                dt = 1e-6
                first_row['cczr'] = -1 * splev(dt, self.cubic_spline_definition, der=0) / dt
            case ZeroCurveInterpMethod.CUBIC_SPLINE_ON_CCZR:
                x = pillar_df['years'].to_list()
                y = pillar_df['cczr'].to_list()
                self.cubic_spline_definition = splrep(x=x, y=y, k=3)
                first_row['cczr'] = splev(0, self.cubic_spline_definition, der=0)
            case _:
                raise ValueError(f"Invalid interpolation method {self.interp_method}")

        first_row_df = pd.DataFrame([first_row], columns=pillar_df.columns)
        pillar_df = pd.concat([first_row_df, pillar_df])
        pillar_df.reset_index(inplace=True, drop=True)

        # Reorder the columns into a consistent format
        column_order = ['tenor','date','days','years','cczr','discount_factor']
        pillar_df = pillar_df[[col for col in column_order if col in pillar_df.columns]]
        
        self.pillar_df = pillar_df
        
        
    def _setup_daily_df(self):
        
        if 'date' in self.pillar_df.columns:
            max_date = max(self.pillar_df['date'])
        else:
            max_years = int(max(self.pillar_df['years']))
            max_date = self.curve_date + relativedelta(years=max_years)
            while year_frac(self.curve_date, max_date, self.day_count_basis) < max_years:
                max_date += dt.timedelta(days=1)
                
        date_range = pd.date_range(self.curve_date,max_date,freq='d')
        days = day_count(self.curve_date, date_range, self.day_count_basis)
        years = year_frac(self.curve_date,date_range, self.day_count_basis)

        if self.interp_method == ZeroCurveInterpMethod.LINEAR_ON_LN_DISCOUNT:
            ln_df_interp = np.interp(x=years, xp=self.pillar_df['years'], fp=np.log(self.pillar_df['discount_factor']))
            cczr = -1 * ln_df_interp / years
            self.daily_df = pd.DataFrame({'date': date_range.to_list(),
                                          'days': days,
                                          'years': years,
                                          'cczr': cczr,
                                          'discount_factor': np.exp(ln_df_interp)})

        elif self.interp_method == ZeroCurveInterpMethod.CUBIC_SPLINE_ON_LN_DISCOUNT:
            ln_df_interp = splev(x=years, tck=self.cubic_spline_definition, der=0)
            cczr = -1 * ln_df_interp / years
            self.daily_df = pd.DataFrame({'date': date_range.to_list(),
                                          'days': days,
                                          'years': years,
                                          'cczr': cczr,
                                          'discount_factor': np.exp(ln_df_interp)})

        elif self.interp_method == ZeroCurveInterpMethod.CUBIC_SPLINE_ON_CCZR:
                zero_rate_interp = splev(years, self.cubic_spline_definition, der=0)
                self.daily_df = pd.DataFrame({'date': date_range.to_list(),
                                              'days': days,
                                              'years': years,
                                              'cczr': zero_rate_interp,
                                              'discount_factor': np.exp(-1 * zero_rate_interp * years)})
        else:
            raise ValueError(f"Invalid interpolation method {self.interp_method}")

       
    def flat_shift(self, 
                   basis_points: float=1) -> 'ZeroCurve':
        dates = list(self.pillar_df.keys())
        years = year_frac(self.curve_date, dates, self.day_count_basis)
        shifted = - np.log(list(self.pillar_df.values)) / years + basis_points / 10000
        shifted = pd.DataFrame({'date': dates,
                                'years': years,
                                'discount_factor':np.exp(-shifted * years)})
        shifted.loc[0,'discount_factor'] = 1.0 # set the discount factor on the curve date to be 1.0
        return ZeroCurve(curve_date=self.curve_date, 
                         pillar_df=shifted,
                         day_count_basis=self.day_count_basis)
                   
    def get_forward_rates(self,
                          period_start: Union[pd.Timestamp, np.datetime64, dt.datetime, dt.date, pd.Series, pd.DatetimeIndex],
                          period_end: Union[pd.Timestamp, np.datetime64, dt.datetime, dt.date, pd.Series, pd.DatetimeIndex],
                          forward_rate_type: [TermRate, RFRFixingCalcMethod]) -> np.array:

        period_start = pd.to_datetime(period_start)
        period_end = pd.to_datetime(period_end)

        if isinstance(period_start, (pd.Series, pd.DatetimeIndex)):
            period_start = pd.DatetimeIndex(period_start)
        else:
            period_start = pd.DatetimeIndex([period_start])

        if isinstance(period_end, (pd.Series, pd.DatetimeIndex)):
            period_end = pd.DatetimeIndex(period_end)
        else:
            period_end = pd.DatetimeIndex([period_end])


        assert len(period_start) == len(period_end)
        assert (period_start >= self.curve_date).all()
        assert (period_end >= period_start).all()

        if forward_rate_type in {RFRFixingCalcMethod.WEIGHTED_AVERAGE, RFRFixingCalcMethod.SIMPLE_AVERAGE}:
            
            dates = pd.date_range(start=period_start.min(), end=period_end.max(), freq='D')
            discount_factors = self.get_discount_factors(dates=dates)
            daily_interest_multiplier = discount_factors[:-1] / discount_factors[1:]
            daily_simple_interest_rate = (daily_interest_multiplier - 1) * self.day_count_basis.days_per_year
            dates_np = dates.to_numpy(dtype='datetime64[D]')
            busday_flag = pd.Series(np.is_busday(dates_np, busdaycal=self.cal), index=dates)
            
            helper_data = {
                    'date': dates[:-1],
                    'daily_interest_multiplier': daily_interest_multiplier,
                    'simple_daily_rate': daily_simple_interest_rate,
                    'business_day_flag': busday_flag.values[:-1]
                }
            helper_df = pd.DataFrame(helper_data)
            
            result = np.full(period_start.shape, np.nan)

            for i,(start_date,end_date) in enumerate(zip(period_start, period_end)):
                mask = np.logical_and(helper_df['date'] >= start_date,
                                      helper_df['date'] < end_date)
                match forward_rate_type:
                    case RFRFixingCalcMethod.WEIGHTED_AVERAGE:
                        result[i] = helper_df.loc[mask,'simple_daily_rate'].mean()
                    case RFRFixingCalcMethod.SIMPLE_AVERAGE:
                        mask = np.logical_and(mask, helper_df['business_day_flag'])
                        result[i] = helper_df.loc[mask,'simple_daily_rate'].mean()
                        
                    # Returns same result as ForwardRate.SIMPLE formulae
                    # case ForwardRate.DAILY_COMPOUNDED:
                    #     year_frac_ = year_frac(start_date, end_date, self.day_count_basis)
                    #     result[i] = (helper_df.loc[mask,'daily_interest_multiplier'].product() - 1) / year_frac_
                        
            return result
        
        else:
            Δt = pd.Series(year_frac(period_start, period_end, self.day_count_basis))
            DF_t1 = self.get_discount_factors(dates=period_start)
            DF_t2 = self.get_discount_factors(dates=period_end)
        
            # https://en.wikipedia.org/wiki/Forward_rate
            if forward_rate_type in {TermRate.SIMPLE, RFRFixingCalcMethod.DAILY_COMPOUNDED}:
                result = (1.0 / Δt) * (DF_t1 / DF_t2 - 1.0)
            elif forward_rate_type == TermRate.CONTINUOUS:
                result = (1.0 / Δt) * (np.log(DF_t1) - np.log(DF_t2))
            elif forward_rate_type == TermRate.ANNUAL:
                result = (DF_t1 / DF_t2) ** (1.0 / Δt)  - 1.0
            else:
                raise ValueError(f"Invalid forward_rate_type {forward_rate_type}")        

            return result.values
        
    def get_instantaneous_forward_rate(self, years):
        if self.interp_method == ZeroCurveInterpMethod.CUBIC_SPLINE_ON_CCZR:
            zero_rate = self.get_zero_rates(years=years, compounding_freq=CompoundingFreq.CONTINUOUS)
            zero_rate_1st_deriv = splev(x=years, tck=self.cubic_spline_definition, der=1)
            return zero_rate + years * zero_rate_1st_deriv
        elif self.interp_method == ZeroCurveInterpMethod.CUBIC_SPLINE_ON_LN_DISCOUNT:
            ln_discount_1st_deriv = splev(x=years, tck=self.cubic_spline_definition, der=1)
            return -ln_discount_1st_deriv


    def get_discount_factors(self,
                            dates: Optional[Union[pd.Timestamp, np.datetime64, dt.datetime, dt.date, pd.Series, pd.DatetimeIndex]]=None,
                            days: Optional[Union[int, pd.Series]]=None,
                            years: Optional[Union[float, pd.Series]]=None) -> pd.Series:

        df = self.index_daily_data(dates, days, years)
        return df['discount_factor'].values
           
        
    def get_zero_rates(self,
                      compounding_freq: CompoundingFreq,
                      dates: Optional[Union[pd.Timestamp, np.datetime64, dt.datetime, dt.date, pd.Series, pd.DatetimeIndex]]=None,
                      days:  Optional[Union[int, pd.Series]]=None,
                      years: Optional[Union[float, pd.Series]]=None
                      ) -> pd.Series:
                
        df = self.index_daily_data(dates, days, years)

        if compounding_freq == CompoundingFreq.CONTINUOUS:
            return df['cczr'].values
        else:
            zero_rate = zero_rate_from_discount_factor(years=df['years'].values,
                                                       discount_factor=df['discount_factor'].values,
                                                       compounding_freq=compounding_freq)
            return zero_rate
        
        
    def index_daily_data(self, 
                  dates: Optional[Union[pd.Timestamp, np.datetime64, dt.datetime, dt.date, pd.Series, pd.DatetimeIndex]]=None,
                  days: Optional[Union[int, pd.Series]]=None,
                  years: Optional[Union[float, pd.Series]]=None) -> pd.DataFrame:
        
        inputs = {'dates': dates, 'days': days, 'years': years}
        if sum(x is not None for x in inputs.values()) != 1:
            raise ValueError('Only one input among days, date, or years is allowed.')            
        
        if years is not None:
            if self.interp_method == 'linear_on_ln_discount':
                ln_df_interp = np.interp(x=years, xp=self.pillar_df['years'], fp= np.log(self.pillar_df['discount_factor']))
                cczr = -1 * ln_df_interp / years
                discount_factor = np.exp(ln_df_interp)
            elif self.interp_method == ZeroCurveInterpMethod.CUBIC_SPLINE_ON_LN_DISCOUNT:
                ln_df_interp = splev(years, self.cubic_spline_definition, der=0)
                cczr = -1 * ln_df_interp / years
                discount_factor = np.exp(ln_df_interp)
            elif self.interp_method == ZeroCurveInterpMethod.CUBIC_SPLINE_ON_CCZR:
                cczr = splev(years, self.cubic_spline_definition, der=0)
                discount_factor = np.exp(-1 * cczr * years)
            else:
                raise ValueError

            years, cczr, discount_factor = map(np.atleast_1d, (years, cczr, discount_factor))
            return pd.DataFrame({'years': years, 'cczr': cczr, 'discount_factor': discount_factor})
        else:
            if days is not None:
                dates = self.curve_date + dt.timedelta(days=days)

            dates = pd.to_datetime(dates)
            if not(isinstance(dates, pd.DatetimeIndex)):
                if isinstance(dates, pd.Series):
                    dates = pd.DatetimeIndex(dates)
                else:
                    dates = pd.DatetimeIndex([dates])

            # Message suffix if any dates are outside the available data
            if self.extrap_method == 'none':
                msg = f". NaN will be returned as 'extrapolation_method' is {self.extrap_method}"
            elif self.extrap_method == 'flat':
                msg = f". Flat extrapolation will be applied as 'extrap_method' is {self.extrap_method}"
            
            # Check if any of the dates are outside the available data
            min_date = self.daily_df['date'].min()
            below_range = dates < min_date
            dates_below = []
            if any(below_range):
                bound_date = self.daily_df['date'].min()
                for i,date in enumerate(dates[below_range]):
                    print('Date', date.strftime('%Y-%m-%d'), 'is below the min available data of ' + bound_date.strftime('%Y-%m-%d') + msg)
                    if self.extrap_method == 'none':
                        dates_below.append(np.nan)
                    elif self.extrap_method == 'flat':
                        dates_below.append(bound_date)
            
            max_date = self.daily_df['date'].max()
            above_range = dates > max_date
            dates_above = []
            if any(above_range):
                bound_date = self.daily_df['date'].max()
                for i,date in enumerate(dates[above_range]):
                    print('Date', date.strftime('%Y-%m-%d'), 'is above the max available data of ' + bound_date.strftime('%Y-%m-%d') + msg)
                    if self.extrap_method == 'none':
                        dates_above.append(np.nan)
                    elif self.extrap_method == 'flat':
                        dates_above.append(bound_date)
          
            in_range = np.logical_and(np.logical_not(below_range),np.logical_not(above_range))
            cleaned_dates = dates_below + list(dates[in_range].values) + dates_above
            df = pd.merge(pd.DataFrame(data=cleaned_dates, columns=['date']), self.daily_df, left_on='date',right_on='date', how='left')
            return df    

       
    def plot(self, forward_rate_term=pd.DateOffset(months=3)):
            
        # Zero rates
        min_date = self.daily_df['date'].iloc[1]
        max_date = self.daily_df['date'].iloc[-1]
        date_range = pd.date_range(min_date,max_date,freq='d')
        years_zr = year_frac(self.curve_date, date_range, self.day_count_basis)
        zero_rates = pd.Series(self.get_zero_rates(compounding_freq=CompoundingFreq.CONTINUOUS, dates=date_range)) * 100
        
        fig, ax = plt.subplots()

        # Zero Rate
        ax.plot(years_zr, zero_rates, label='zero rate')  
  
        # Forward rates
        d1 = pd.date_range(min_date,max_date - forward_rate_term,freq='d')
        d2 = pd.date_range(min_date + forward_rate_term,max_date ,freq='d')
        years_fwd = year_frac(self.curve_date, d1, self.day_count_basis)
        fwd_rates = pd.Series(self.get_forward_rates(d1, d2, TermRate.SIMPLE)) * 100
        ax.plot(years_fwd, fwd_rates)
        
        ax.set(xlabel='years', ylabel='interest rate (%)')
        
        ax.grid()
        ax.legend()
        plt.xlim([min(years_zr), 1.05*max(years_zr)])
        plt.show()



        
        

        
        
    


# -*- coding: utf-8 -*-
import os
from dataclasses import dataclass, field
import numpy as np
import pandas as pd
from typing import List, Tuple, Union
import datetime
from frm.enums.utils import RollConv, TimingConvention, Stub, PeriodFreq, DayRoll, DayCountBasis, ExchangeNotionals
from frm.utils.daycount import year_frac, day_count


if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))


@dataclass
class BaseSchedule:
    # Base Schedule parameters
    start_date: [pd.Timestamp, np.datetime64, datetime.date, datetime.datetime]
    end_date: [pd.Timestamp, np.datetime64, datetime.date, datetime.datetime]
    freq: PeriodFreq
    roll_conv: RollConv = RollConv.MODIFIED_FOLLOWING
    day_roll: DayRoll = DayRoll.UNADJUSTED
    first_period_end: Union[pd.Timestamp, np.datetime64, datetime.date, datetime.datetime, None] = None
    last_period_start: Union[pd.Timestamp, np.datetime64, datetime.date, datetime.datetime, None] = None
    first_stub: Stub = Stub.DEFAULT
    last_stub: Stub = Stub.DEFAULT
    roll_user_specified_dates: bool = False
    cal: np.busdaycalendar = np.busdaycalendar()
    # The schedule DataFrame is initialized after the object is created
    df: pd.DataFrame = field(init=False)

    def __post_init__(self):
        # Convert to pandas Timestamp
        self.start_date = pd.Timestamp(self.start_date)
        self.end_date = pd.Timestamp(self.end_date)
        self.first_period_end = pd.Timestamp(self.first_period_end) if self.first_period_end is not None else None
        self.last_period_start = pd.Timestamp(self.last_period_start) if self.last_period_start is not None else None
        self.make_schedule()

    def make_schedule(self):
        # Generate the schedule DataFrame using the stored parameters
        self.df = make_schedule(
            start_date=self.start_date,
            end_date=self.end_date,
            freq=self.freq,
            roll_conv=self.roll_conv,
            day_roll=self.day_roll,
            first_period_end=self.first_period_end,
            last_period_start=self.last_period_start,
            first_stub=self.first_stub,
            last_stub=self.last_stub,
            roll_user_specified_dates=self.roll_user_specified_dates,
            cal=self.cal
        )

        if self.roll_user_specified_dates:
            self.start_date = self.df['period_start'].iloc[0]
            self.end_date = self.df['period_end'].iloc[-1]

    def add_payment_dates(self,
                          payment_delay: int=0,
                          payment_timing: TimingConvention = TimingConvention.IN_ARREARS,
                          col_name: str='payment_date'):
        """
        Add payment dates to the schedule DataFrame for valid periods.

        Parameters
        ----------
        payment_delay : int
            Specifies how many days after period start_date/end_date (if payments are in_advance/in_arrears), the payment is made. The default is 0.
        payment_timing : TimingConvention
            Specifies when payments are made. The default is TimingConvention.IN_ARREARS.
        col_name: str
            Column name to use for the payment dates. The default is 'payment_date'.
        """
        match payment_timing:
            case TimingConvention.IN_ARREARS:
                dates = self.df['period_end'].to_numpy(dtype='datetime64[D]')
            case TimingConvention.IN_ADVANCE:
                dates = self.df['period_start'].to_numpy(dtype='datetime64[D]')
            case _:
                raise ValueError(f"Invalid payment_timing {payment_timing}")

        # Convert NaT values to a mask
        valid_dates_mask = ~pd.isna(dates)
        valid_dates = dates[valid_dates_mask]

        # Apply np.busday_offset only to non-NaT dates
        adjusted_dates = np.empty_like(dates, dtype='datetime64[D]')
        adjusted_dates[valid_dates_mask] = np.busday_offset(valid_dates, offsets=payment_delay, roll=self.roll_conv.value, busdaycal=self.cal)
        adjusted_dates[~valid_dates_mask] = 'NaT'

        # Update DataFrame
        self.df[col_name] = pd.DatetimeIndex(adjusted_dates).astype('datetime64[ns]')

    def add_period_daycount(self, day_count_basis: DayCountBasis):
        """ Add the period length in days to the schedule DataFrame, always replacing existing columns"""

        # Remove 'period_days' and 'period_years' if they already exist
        if 'period_daycount' in self.df.columns:
            self.df.pop('period_daycount')

        # Get the column index for 'period_end'
        col_index = self.df.columns.get_loc('period_end')

        # Insert 'period_days' and 'period_years' at the specified positions
        mask = np.logical_and(self.df['period_start'].notnull(), self.df['period_end'].notnull())
        days = np.full(self.df.shape[0], np.nan)
        days[mask] = day_count(self.df['period_start'][mask], self.df['period_end'][mask], day_count_basis)
        self.df.insert(loc=col_index + 1, column='period_daycount',value=days)

    def add_period_yearfrac(self, day_count_basis: DayCountBasis):
        """ Add the period length years to the schedule DataFrame, always replacing existing columns"""

        if 'period_yearfrac' in self.df.columns:
            self.df.pop('period_yearfrac')

        # Get the column index for 'period_end'
        col_index = self.df.columns.get_loc('period_end')

        # Insert 'period_days' and 'period_years' at the specified positions
        mask = np.logical_and(self.df['period_start'].notnull(), self.df['period_end'].notnull())
        years = np.full(self.df.shape[0], np.nan)
        years[mask] = year_frac(self.df['period_start'][mask], self.df['period_end'][mask], day_count_basis)
        self.df.insert(loc=col_index + 2, column='period_yearfrac', value=years)


@dataclass
class NotionalSchedule(BaseSchedule):
    notional_amount: float | np.ndarray = 100_000_000
    exchange_notionals: ExchangeNotionals = ExchangeNotionals.NEITHER
    notional_payment_delay: int = 0
    notional_payment_timing: TimingConvention = TimingConvention.IN_ARREARS
    initial_notional_exchange_date: Union[pd.Timestamp, np.datetime64, datetime.date, datetime.datetime, None] = None

    def __post_init__(self):
        super().__post_init__()
        self.add_notional_schedule()

    def add_notional_schedule(self):
        """Adds columns 'notional', 'notional_payment' and 'notional_payment_date' to the schedule DataFrame"""

        notional_amount = np.atleast_1d(self.notional_amount)
        if notional_amount.shape == (1,):
            self.df['notional'] = notional_amount[0]
            self.df['notional_payment'] = 0
        elif notional_amount.shape == self.df.shape[0]:
            self.df['notional'] = notional_amount
            self.df['notional_payment'] = notional_amount[:-1] - notional_amount[1:]
        else:
            raise ValueError("Invalid notional_amount shape")

        self.add_payment_dates(self.notional_payment_delay, self.notional_payment_timing, col_name='notional_payment_date')

        if self.exchange_notionals == ExchangeNotionals.START or self.exchange_notionals == ExchangeNotionals.BOTH:
            column_order = self.df.columns
            if self.initial_notional_exchange_date is None:
                self.initial_notional_exchange_date = self.df['period_start'].iloc[0]

            row_data = {'notional_payment_date': self.initial_notional_exchange_date}
            self.df = pd.concat([pd.DataFrame(row_data, index=[0]), self.df], ignore_index=True)
            self.df = self.df[column_order]
            self.df.reset_index(drop=True, inplace=True)

            self.df.loc[self.df.index[0], 'notional'] = 0
            self.df.loc[self.df.index[0], 'notional_payment'] = -1 * notional_amount[0]

        if self.exchange_notionals == ExchangeNotionals.END or self.exchange_notionals == ExchangeNotionals.BOTH:
            # Add final notional exchange period
            self.df.loc[self.df.index[-1], 'notional_payment'] = notional_amount[-1]


@dataclass
class CouponSchedule(NotionalSchedule):
    coupon_payment_delay: int = 0
    coupon_payment_timing: TimingConvention = TimingConvention.IN_ARREARS

    def __post_init__(self):
        super().__post_init__()
        self.add_payment_dates(self.coupon_payment_delay, self.coupon_payment_timing, col_name='coupon_payment_date')

    def determine_valid_shapes_for_coupon_param(self):
        """Determine valid shapes for the contractual coupon param (fixed rate, spread) based on exchange_notionals."""
        row_count = self.df.shape[0]
        if self.exchange_notionals in [ExchangeNotionals.START, ExchangeNotionals.BOTH]:
            index = self.df.index[1:] # The first row is the initial exchange of notionals, with no coupon
            valid_shape = [(1,), (row_count - 1,)]
        else:
            index = self.df.index[0:]
            valid_shape = [(1,), (row_count,)]
        return index, valid_shape






def make_schedule(
        start_date: [pd.Timestamp, np.datetime64, datetime.date, datetime.datetime],
        end_date: [pd.Timestamp, np.datetime64, datetime.date, datetime.datetime],
        freq: PeriodFreq,
        roll_conv: RollConv=RollConv.MODIFIED_FOLLOWING,
        day_roll: DayRoll=DayRoll.UNADJUSTED,
        first_period_end: Union[pd.Timestamp, np.datetime64, datetime.date, datetime.datetime, None] = None,
        last_period_start: Union[pd.Timestamp, np.datetime64, datetime.date, datetime.datetime, None] = None,
        first_stub: Stub=Stub.DEFAULT,
        last_stub: Stub=Stub.DEFAULT,
        roll_user_specified_dates: bool=False,
        cal: np.busdaycalendar=np.busdaycalendar(),
        ) -> pd.DataFrame:
    """
    Create a schedule. Optional detailed stub logic.

    Parameters
    ----------
    start_date : pandas.Timestamp
        Specifies the effective date of the schedule
    end_date : pandas.Timestamp
        Specifies the termination date of the schedule
    freq : str
        Specify the period frequency
    roll_conv : RollConv
        How to treat dates that do not fall on a valid day. The default is RollConv.MODIFIED_FOLLOWING.
    day_roll : DayRoll
        Specifies the day periods should start/end on. The default is DayRoll.UNADJUSTED.
    first_period_end: pandas.Timestamp
        Specifies the end date of the first period. The first_period_end overrides the first_stub field.
    last_period_start: pandas.Timestamp
        Specifies the start date of the last period. The last_period_start overrides the last_stub field.
    first_stub : Stub
        Specifies the type of the first stub. If first_period_end is specified, the first_stub is ignored.
    last_stub : Stub
        Specifies the type of the last stub. If last_period_start is specified, the last_stub is ignored.
    roll_user_specified_dates : bool
        Boolean flag for whether to roll (per business day calendar and roll convention) the user specified dates (start_date, end_date, first_period_end, last_period_start)
    cal : np.busdaycalendar
        Specifies the business day calendar to observe.

    Returns
    -------
    schedule : pandas.DataFrame
        Columns:
            - fixing_date (if add_fixing_dates=True)
            - period_start
            - period_end
            - payment_date (if add_payment_dates=True)
    """

    start_date, end_date, first_period_end, last_period_start = [
        pd.Timestamp(d) if isinstance(d, (np.datetime64, datetime.date, datetime.datetime)) else d
        for d in [start_date, end_date, first_period_end, last_period_start]
    ]

    # Input validation
    if start_date >= end_date:
        raise ValueError(f"start_date {start_date} must be before end_date {end_date}")

    if first_period_end is not None:
        assert start_date < first_period_end <= end_date

    if last_period_start is not None:
        assert start_date <= last_period_start < end_date

    # if freq == PeriodFreq.CDS:
    #     pass
    # elif freq == PeriodFreq.IMM:
    #     pass

    if freq == PeriodFreq.ZERO_COUPON or first_period_end == end_date or last_period_start == start_date:
        # Function inputs explicitly specify 1 period
        d1, d2 = [start_date], [end_date]
    elif first_period_end == last_period_start and first_period_end is not None and last_period_start is not None:
        # Function inputs explicitly specify 2 periods
        d1 = [start_date, first_period_end]
        d2 = [last_period_start, end_date]
    else:
        if first_period_end is None and last_period_start is None:
            assert first_stub != Stub.DEFINED_PER_FIRST_PERIOD_END_DATE
            assert last_stub != Stub.DEFINED_PER_LAST_PERIOD_START_DATE

            if first_stub == Stub.DEFAULT and last_stub == Stub.DEFAULT:
                # If no stub is specified, defaults to market convention on the 1st stub, no last stub.
                first_stub = Stub.market_convention()
                last_stub = Stub.NONE
            elif first_stub == Stub.DEFAULT:
                first_stub  = Stub.NONE
            elif last_stub == Stub.DEFAULT:
                last_stub = Stub.NONE
            else:
                raise ValueError("Unexpected logic branch - please raise GitHub issue")

            if last_stub == Stub.NONE:
                d1, d2 = generate_date_schedule(start_date, end_date, freq, 'backward', roll_conv, day_roll, roll_user_specified_dates, cal)
                if first_stub == Stub.LONG:
                    d1 = [d1[0]] + d1[2:]
                    d2 = d2[1:]
            elif first_stub == Stub.NONE and last_stub != Stub.NONE:
                d1, d2 = generate_date_schedule(start_date, end_date, freq, 'forward', roll_conv, day_roll, roll_user_specified_dates, cal)
                if last_stub == Stub.LONG:
                    d1 = d1[:-1]
                    d2 = d2[:-2] + [d2[-1]]
            elif first_stub in [Stub.SHORT, Stub.LONG] and last_stub in [Stub.SHORT, Stub.LONG]:
                raise ValueError("If a schedule has first and last stubs they must be specified via first_period_end and last_period_start")
            else:
                raise ValueError("Unexpected logic branch - please raise GitHub issue")

        else:
            # If first_period_end or last_period_start are specified we want to generate the schedule
            # via generate_date_schedule(start_date, end_date, ...) and see if
            # the first_period_end or last_period_start match the generated schedule.
            #
            # If they align, we use this generated schedule.
            #
            # If they don't align, we generate the schedule:
            # (i) via backward generation from last_period_start if only last_period_start is specified
            # (ii) via forward generation from first_period_end if only first_period_end is specified
            # (iii) between first_period_end and last_period_start if both are specified

            # Step 1: Generate the schedule by backward and forward date generation
            d1_backward, d2_backward = generate_date_schedule(start_date, end_date, freq, 'backward', roll_conv, day_roll, roll_user_specified_dates, cal)
            d1_forward, d2_forward = generate_date_schedule(start_date, end_date, freq, 'forward', roll_conv, day_roll, roll_user_specified_dates, cal)

            # Step 2a : Check if the first_period_end matches any generated schedules

            direction = []
            if first_period_end is not None:
                if first_period_end == d2_forward[0]:
                    direction.append('forward')
                    first_stub = Stub.NONE
                elif first_period_end == d2_backward[0]:
                    direction.append('backward')
                    first_stub = Stub.SHORT
                elif first_period_end == d2_backward[1]:
                    direction.append('backward')
                    first_stub = Stub.LONG
                else:
                    # Need to construct schedule using first_period_end
                    first_stub = Stub.DEFINED_PER_FIRST_PERIOD_END_DATE

            # Step 2a : Check if the last_period_start matches any generated schedules
            if last_period_start is not None:
                if last_period_start == d1_backward[-1]:
                    direction.append('backward')
                    last_stub = Stub.NONE
                elif last_period_start == d1_forward[-1]:
                    direction.append('forward')
                    last_stub = Stub.SHORT
                elif last_period_start == d1_forward[-2]:
                    direction.append('forward')
                    last_stub = Stub.LONG
                else:
                    # Need to construct schedule using last_period_start
                    last_stub = Stub.DEFINED_PER_LAST_PERIOD_START_DATE

            # Step 2c : Set default stub values to:
            #              (i) NONE if other stub is SHORT/LONG
            #              (ii) SHORT if other stub is NONE
            if first_stub == Stub.DEFAULT and last_stub != Stub.DEFINED_PER_LAST_PERIOD_START_DATE:
                assert last_stub != Stub.DEFAULT
                if last_stub == Stub.NONE:
                    first_stub = Stub.market_convention()
                else:
                    first_stub = Stub.NONE
            if last_stub == Stub.DEFAULT and first_stub != Stub.DEFINED_PER_FIRST_PERIOD_END_DATE:
                assert first_stub != Stub.DEFAULT
                if first_stub == Stub.NONE:
                    last_stub = Stub.market_convention()
                else:
                    last_stub = Stub.NONE

            # Step 3: Construct the schedules
            if first_stub == Stub.DEFINED_PER_FIRST_PERIOD_END_DATE and last_stub == Stub.DEFINED_PER_LAST_PERIOD_START_DATE:
                # Generate date schedule between first_period_end and last_period_start
                d1, d2 = generate_date_schedule(first_period_end, last_period_start, freq, 'backward', roll_conv, day_roll, roll_user_specified_dates, cal)
                d1 = [start_date] + d1 + [last_period_start]
                d2 = [first_period_end] + d2 + [end_date]
            elif first_stub == Stub.DEFINED_PER_FIRST_PERIOD_END_DATE and last_stub != Stub.DEFINED_PER_LAST_PERIOD_START_DATE:
                # Forward generation from first_period_end and if long last stub, combine last two periods
                d1, d2 = generate_date_schedule(first_period_end, end_date, freq, 'forward', roll_conv, day_roll, roll_user_specified_dates, cal)
                d1 = [start_date] + d1
                d2 = [first_period_end] + d2
                if last_stub == Stub.LONG:
                    d1 = d1[:-1]
                    d2 = d2[:-2] + [d2[-1]]
            elif first_stub != Stub.DEFINED_PER_FIRST_PERIOD_END_DATE and last_stub == Stub.DEFINED_PER_LAST_PERIOD_START_DATE:
                # Backward generation from last_period_start and if long first stub, combine first two periods
                d1, d2 = generate_date_schedule(start_date, last_period_start, freq, 'backward', roll_conv, day_roll, roll_user_specified_dates, cal)
                d1 = d1 + [last_period_start]
                d2 = d2 + [end_date]
                if first_stub == Stub.LONG:
                    d1 = [d1[0]] + d1[2:]
                    d2 = d2[1:]
            else:
                # Don't need to generate any additional schedules
                assert len(set(direction)) == 1

                if first_stub == Stub.NONE:
                    # Forward date generation
                    d1, d2 = d1_forward, d2_forward
                    if last_stub == Stub.LONG:
                        d1 = d1[:-1]
                        d2 = d2[:-2] + [d2[-1]]
                elif last_stub == Stub.NONE:
                    # Backward date generation
                    d1, d2 = d1_backward, d2_backward
                    if first_stub == Stub.LONG:
                        d1 = [d1[0]] + d1[2:]
                        d2 = d2[1:]
                else:
                    raise ValueError("Unexpected logic branch - please raise GitHub issue")

    schedule = pd.DataFrame({'period_start': d1, 'period_end': d2})
    return schedule


def generate_date_schedule(
        start_date: pd.Timestamp, 
        end_date: pd.Timestamp, 
        freq: PeriodFreq,
        direction: str,
        roll_conv: RollConv=RollConv.MODIFIED_FOLLOWING,
        day_roll: DayRoll=DayRoll.UNADJUSTED,
        roll_user_specified_dates: bool=False,
        cal: np.busdaycalendar=np.busdaycalendar()
    ) -> Tuple[List, List]:
    """
    Generates a schedule of start and end dates between start_date and end_date.
    
    Parameters
    ----------
    start_date : pd.Timestamp
        The start date of the schedule.
    end_date : pd.Timestamp
        The end date of the schedule.
    freq : PeriodFreq
        The frequency of the schedule.
    direction : {'forward', 'backward'}
        The direction in which to generate dates.
    roll_conv : RollConv
        How to treat dates that fall on a non business day.
    day_roll : DayRoll
        Specifies the day periods should start/end on.
    roll_user_specified_dates : bool
        Boolean flag for whether to roll (per business day calendar and roll convention) the user specified dates (start_date, end_date) 
    cal : np.busdaycalendar
        Specifies the business day calendar to observe.

    Returns
    -------
    Tuple[List, List]
    
    Raises
    ------
    ValueError, TypeError
        If any of the inputs have invalid values or types

    TODO: Consider options for passing param=(start_date, end_date, freq, None, None, None, None, None, None, None, None) to function
          E.g. Wrap with fields with Optional[] and set defaults in function from enum.set_default()
    """
    
    def busday_offset_timestamp(pd_timestamp, offsets, roll_conv, cal):
        np_datetime64D = np.array([pd_timestamp]).astype('datetime64[D]')
        rolled_date_np = np.busday_offset(np_datetime64D, offsets=offsets, roll=roll_conv, busdaycal=cal)[0]
        return pd.Timestamp(rolled_date_np)
    
    def apply_specific_day_roll(pd_timestamp: pd.Timestamp,
                                specific_day_roll: DayRoll) -> pd.Timestamp:
        if specific_day_roll == DayRoll.UNADJUSTED:
            return pd_timestamp
        else:
            try:
                return pd_timestamp.replace(day=specific_day_roll.value)
            except ValueError:
                return pd_timestamp + pd.offsets.MonthEnd(0)

    # Input validation
    if direction not in {'forward', 'backward'}:
        raise ValueError(f"Invalid direction '{direction}'. Must be 'forward' or 'backward'.")
    if not isinstance(start_date, pd.Timestamp) or not isinstance(end_date, pd.Timestamp):
        raise TypeError("'start_date' {start_date} and 'end_date' {end_date} must be pandas Timestamp objects.")
    if start_date >= end_date:
        raise ValueError("'start_date' must be earlier than 'end_date'.")    
        
    # Implementation rationale:
    # - append to both start_dates and end_dates in while loop so don't have to consider edge case of list length = 1
    # - Note the 'roll day' component must be applied inside the while loop due to holidays/weekends 
    #   which could move a day past the while loop condition.
    i = 1
    start_dates = []
    end_dates = []
    if roll_user_specified_dates:
        start_date_in_schedule = busday_offset_timestamp(start_date, 0, roll_conv.value, cal)
        end_date_in_schedule = busday_offset_timestamp(end_date, 0, roll_conv.value, cal)
    else:
        start_date_in_schedule = start_date
        end_date_in_schedule = end_date

    if direction == 'forward':
        current_date = apply_specific_day_roll(start_date + freq.date_offset, day_roll)
        current_date = busday_offset_timestamp(current_date, 0, roll_conv.value, cal)
        
        while current_date < end_date: 
            start_dates.append(current_date)
            end_dates.append(current_date)
            current_date = start_date + freq.multiply_date_offset(i+1)
            current_date = apply_specific_day_roll(current_date, day_roll)
            current_date = busday_offset_timestamp(current_date, 0, roll_conv.value, cal)
            i += 1

        start_dates = [start_date_in_schedule] + start_dates
        end_dates.append(end_date_in_schedule)
        
    elif direction == 'backward':
        current_date = apply_specific_day_roll(end_date - freq.date_offset, day_roll)
        current_date = busday_offset_timestamp(current_date, 0, roll_conv.value, cal)
        
        while current_date > start_date:   
            start_dates.append(current_date)
            end_dates.append(current_date)
            current_date = end_date - freq.multiply_date_offset(i+1)
            current_date = apply_specific_day_roll(current_date, day_roll)
            current_date = busday_offset_timestamp(current_date, 0, roll_conv.value, cal)
            i +=1

        start_dates.append(start_date_in_schedule)
        end_dates = [end_date_in_schedule] + end_dates
        
        start_dates.reverse()
        end_dates.reverse()
        
    return start_dates, end_dates
    

def create_date_grid_for_fx_exposures(curve_date: pd.Timestamp, 
                                      delivery_dates: np.array, 
                                      sampling_freq: str=None,
                                      date_grid_pillar: pd.DatetimeIndex=None,
                                      payments_on_value_date_have_value: bool=False,
                                      include_last_day_of_value: bool=False,
                                      include_day_after_last_day_of_value: bool=False) -> pd.DatetimeIndex:
    """
    Defines a date grid for calculating exposures
    Please note, for each trade, this defines the settlement date grid, hence the expiry
    date grid needs to be solved based on the delay of the market.
    For FX derivatives this is relively simple as market data inputs can be interploted from
    (i) expiry or (ii) settle dates.

    Parameters
    ----------
    curve_date : pd.Timestamp
        DESCRIPTION.
    delivery_dates : np.array
        DESCRIPTION.
    sampling_freq : str, optional
        DESCRIPTION. The default is None.
    date_grid_pillar : pd.DatetimeIndex, optional
        DESCRIPTION. The default is None.
    include_payments_on_value_date : bool, optional
        DESCRIPTION. The default is False.
    include_last_day_of_value : bool, optional
        DESCRIPTION. The default is False.
    include_day_after_last_day_of_value : bool, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    date_grid : TYPE
        DESCRIPTION.

    """
        
    # Challenge is that exposures are defined using the rate-fixing - i.e the expiry date
    # The expiry dates are what we need to specify for our FX option valuation or FX rate simulation
    # The trade technically still has value until some time on day of delivery/settlement.
    # Hence the date grid should be constructed on a delivery/settlement basis. 
    # There is a pototo/potata view on if a cashflow has value on the settlement date. 
    # Our default view is no, it should be in accounts receivable, but this can be toggled. 
    
    delivery_dates = np.unique(delivery_dates) 
    max_settlement_date = delivery_dates.max()
    
    if sampling_freq is not None:    
        if sampling_freq == '1d':
            date_grid = generate_date_schedule(curve_date, max_settlement_date, 1, 'days', 'forward')
        elif sampling_freq == '1w':
            date_grid = generate_date_schedule(curve_date, max_settlement_date, 1, 'weeks', 'forward')
        elif sampling_freq == '1m':
            date_grid = generate_date_schedule(curve_date, max_settlement_date, 1, 'months', 'forward')
        elif sampling_freq == '3m':
            date_grid = generate_date_schedule(curve_date, max_settlement_date, 3, 'months', 'forward')    
        elif sampling_freq == '6m':
             date_grid = generate_date_schedule(curve_date, max_settlement_date, 6, 'months', 'forward')
        elif sampling_freq == '12m':
             date_grid = generate_date_schedule(curve_date, max_settlement_date, 12, 'months', 'forward') 
             
        date_grid = pd.DatetimeIndex([curve_date]).append(pd.DatetimeIndex(date_grid[1]))
        
    elif date_grid_pillar is not None:
        assert isinstance(date_grid_pillar, pd.DatetimeIndex) 
        date_grid = date_grid_pillar
    else:
        raise ValueError("Either 'sampling_freq' or date_grid_pillar' must be specified")

    if include_last_day_of_value:
        if payments_on_value_date_have_value:
            last_day_of_value = delivery_dates
        else: 
            last_day_of_value = delivery_dates - pd.DateOffset(days=1)
        date_grid = pd.DatetimeIndex(np.concatenate(date_grid.values, last_day_of_value))

    elif include_day_after_last_day_of_value:
        if payments_on_value_date_have_value:
            last_day_of_value = delivery_dates + pd.DateOffset(days=1)
        else: 
            last_day_of_value = delivery_dates 
        date_grid = pd.DatetimeIndex(np.concatenate(date_grid.values, last_day_of_value))

    date_grid = date_grid.drop_duplicates().sort_values()
    return date_grid




# -*- coding: utf-8 -*-
from abc import ABC
from dataclasses import dataclass
import os
import pandas as pd
from typing import Optional, Union



from frm.instruments.leg import FixedLeg, FloatTermLeg, FloatRFRLeg, ZerocouponLeg

if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))

# Swap class
# Two legs, pay leg and rcv leg.

# Multi curve class
# It is possible to use zero curves + fixings though there will be lots of fiddly bits to get right:
# - DV01 (if swap uses multiple curves of same currency)
# - the kinked AUD swap curve (for swaption term structure construction / IRS bootstrapping
# - converting a swaption vol surface from 3M to 6M using basis spreads.
# - keeping track of basis spread curves.
# - stub coupon calculation - use part of the 1M and 3M curves.
# Hence, in a multi-curve class we would:
# - have helpers for G10 swap curves
# - do global DV01 shift across AONIA, BBSW 1M/3M/6M.
# - bootstrap
# - +/- 100 bps shift to par rates or DV01 done on par rates
# We want to be able to add various rates:

# AONIA
# (a) AONIA swap rates
# (b) AONIA and BBSW 3M basis spreads

# Defining the 3M and 6M curves:
# (a) BBSW 3M swap rates for <= 3Y and BBSW 6M swap rates for > 3Y and basis spreads
# (b) BBSW 6M swap rates and 3M/6M basis spreads
# (c) BBSW 3M swap rates and 3M/6M basis spreads
# (d) BBSW 6M swap rates and BBSW 3M swap rates

# Defining the 1M curve.
# BBSW 1m/3m basis spreads

# Defining the FCB curve
# AONIA/BBSW 3M basis spread curve
# AONIA vs SOFR FCB curve, BBSW 3M vs SOFR FCB curve

# The user defines buckets of quotes:
# The quotes either (i) purely define a zero curve or (ii) define the basis between two curves.
# It is probably best to "hard code" the process.

# Method:
# Choice (a) OIS DC Stripping? Requires to have fixed rates for OIS curve if doing a iterative bootstrap solve.
# For each instrument defining a solve user needs to define
# (i) if the curve is a forward curve only solve
# (ii) or fwd and discount curve solve (i.e. AONIA, or BBSW 3M under pure IBOR curve)
# (iii) or discount curve only solve (for FCB)
# If it is a forward curve solve, the user needs to the discount curve.



@dataclass
class Swap(ABC):
    "Generic class to cover pricing of all types of swap legs"

    # Required parameters
    leg1: Union[FixedLeg, FloatTermLeg, FloatRFRLeg, ZerocouponLeg]
    leg2: Union[FixedLeg, FloatTermLeg, FloatRFRLeg, ZerocouponLeg]

    # Optional parameters. Potentially move into an "Instrument" class.
    trade_date: Optional[pd.Timestamp] = None
    trade_price: Optional[float] = 0
    trade_id: Optional[str] = None
    trade_comment: Optional[str] = None
    # own_cds_curve: Optional[CDSCurve] = None
    # counterparty_cds_curve: Optional[CDSCurve] = None

    def calc_dirty_pv(self):
        leg1_pv = self.leg1.calc_dirty_pv()
        leg2_pv = self.leg2.calc_dirty_pv()
        return leg1_pv + leg2_pv

    def calc_accrued_interest(self):
        leg1_accrued_interest = self.leg1.calc_accrued_interest()
        leg2_accrued_interest = self.leg2.calc_accrued_interest()
        return leg1_accrued_interest + leg2_accrued_interest

    def calc_clean_pv(self):
        return self.calc_dirty_pv() - self.calc_accrued_interest()

    def par_solve_each_leg(self):
        leg1_par = self.leg1.par_solve()
        leg2_par = self.leg2.par_solve()
        return leg1_par, leg2_par

    def target_solve_swap(self, target_pv: float, leg_number: int):
        if leg_number == 1:
            return self.leg1.target_solve(target_pv)
        elif leg_number == 2:
            return self.leg2.target_solve(target_pv)
        else:
            raise ValueError('leg_number must be 1 or 2')

    def __post_init__(self):
        assert self.leg1.pay_rcv != self.leg2.pay_rcv, 'Legs must have opposite pay/receive perspectives'


# Potentially, create a bunch of classes for different types of swaps that fill in default values (i.e. ZC swap, Fixed/Float swap, etc)
# Then, for common quote types, can define all the "non-quote" parameters in the class and then pass in the quote parameters.


# -*- coding: utf-8 -*-
import os
if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))
    
from enum import Enum
import pandas as pd
import numpy as np
from frm.enums.helper import  clean_enum_value, is_valid_enum_value, get_enum_member


class ZeroCurveInterpMethod(Enum):
    LINEAR_ON_LN_DISCOUNT = 'linear_on_ln_discount'
    CUBIC_SPLINE_ON_LN_DISCOUNT = 'cubic_spline_on_ln_discount'
    CUBIC_SPLINE_ON_CCZR = 'cubic_spline_on_cczr'

    @classmethod
    def default(cls):
        return cls.LINEAR_ON_LN_DISCOUNT # Return the default enum value

    @classmethod
    def is_valid(cls, value):
        value = clean_enum_value(value)
        return value in {enum_member.value for enum_member in cls}

    @classmethod
    def from_value(cls, value):
        """Create an enum member from the given value, if valid."""

        def specific_cleaning(value):
            return value

        cleaned_value = clean_enum_value(value)
        cleaned_value = specific_cleaning(cleaned_value)

        if cleaned_value is None:
            return cls.default()
        for enum_member in cls:
            if enum_member.value == cleaned_value \
                    or enum_member.value.replace('/', '') == cleaned_value:
                return enum_member

        # List all valid codes in case of an error
        valid_values = [enum_member.value for enum_member in cls]
        raise ValueError(f"Invalid value: {value}. Valid codes are: {valid_values}")

    @property
    def display_name(self):
        dict_ = {
            'LINEAR_ON_LN_DISCOUNT': 'Linear on log of discount factors',
            'CUBIC_SPLINE_ON_LN_DISCOUNT': 'Cubic spline on log of discount factors',
            'CUBIC_SPLINE_ON_CCZR': 'Cubic spline on continuously compounded zero rates'
        }
        return dict_[self.name]


class ZeroCurveExtrapMethod(Enum):
    NONE = 'none'
    FLAT = 'flat'

    @classmethod
    def default(cls):
        return cls.NONE # Return the default enum value

    @classmethod
    def is_valid(cls, value):
        value = clean_enum_value(value)
        return value in {enum_member.value for enum_member in cls}

    @classmethod
    def from_value(cls, value):
        """Create an enum member from the given value, if valid."""

        def specific_cleaning(value):
            return value

        cleaned_value = clean_enum_value(value)
        cleaned_value = specific_cleaning(cleaned_value)

        if cleaned_value is None:
            return cls.default()
        for enum_member in cls:
            if enum_member.value == cleaned_value \
                    or enum_member.value.replace('/', '') == cleaned_value:
                return enum_member

        # List all valid codes in case of an error
        valid_values = [enum_member.value for enum_member in cls]
        raise ValueError(f"Invalid value: {value}. Valid codes are: {valid_values}")

    @property
    def display_name(self):
        return self.name.title()



class PayRcv(Enum):
    PAY = 'pay'
    RCV = 'rcv'

    @property
    def multiplier(self):
        return -1 if self == PayRcv.PAY else 1

    @classmethod
    def is_valid(cls, value):
        value = clean_enum_value(value)
        return value in {enum_member.value for enum_member in cls}

    @classmethod
    def from_value(cls, value):
        """Create an enum member from the given value, if valid."""

        def specific_cleaning(value):
            return value

        cleaned_value = clean_enum_value(value)
        cleaned_value = specific_cleaning(cleaned_value)

        if cleaned_value is None:
            return cls.default()
        for enum_member in cls:
            if enum_member.value == cleaned_value \
                    or enum_member.value.replace('/', '') == cleaned_value:
                return enum_member

        # List all valid codes in case of an error
        valid_values = [enum_member.value for enum_member in cls]
        raise ValueError(f"Invalid value: {value}. Valid codes are: {valid_values}")

    @property
    def display_name(self):
        return self.name.title()


class BuySell(Enum):
    BUY = 'buy'
    SELL = 'sell'

    @property
    def multiplier(self):
        return -1 if self == BuySell.SELL else 1

    @classmethod
    def is_valid(cls, value):
        value = clean_enum_value(value)
        return value in {enum_member.value for enum_member in cls}

    @classmethod
    def from_value(cls, value):
        """Create an enum member from the given value, if valid."""

        def specific_cleaning(value):
            return value

        cleaned_value = clean_enum_value(value)
        cleaned_value = specific_cleaning(cleaned_value)

        if cleaned_value is None:
            return cls.default()
        for enum_member in cls:
            if enum_member.value == cleaned_value \
                    or enum_member.value.replace('/', '') == cleaned_value:
                return enum_member

        # List all valid codes in case of an error
        valid_values = [enum_member.value for enum_member in cls]
        raise ValueError(f"Invalid value: {value}. Valid codes are: {valid_values}")

    @property
    def display_name(self):
        return self.name.title()


class DayCountBasis(Enum):
    # If storing instrument definitions in CDM, use 'code' to define valid fieldnames
    _30_360 = '30/360'
    _30E_360 = '30e/360'
    _30E_360_ISDA = '30e/360_isda'
    ACT_360 = 'act/360'
    ACT_365 = 'act/365'
    ACT_ACT = 'act/act'
    ACT_366 = 'act/366'

    def __init__(self, value):        
        days_per_year = {
            '30/360': 360,
            '30e/360': 360,
            '30e/360_isda': 360,
            'act/360': 360,
            'act/365': 365,
            'act/act': np.nan,
            'act/366': 366,
            }
        self.days_per_year = days_per_year[self.value]

    @classmethod
    def default(cls):
        return cls.ACT_ACT # Return the default enum value

    @classmethod
    def is_valid(cls, value):
        value = clean_enum_value(value)
        return value in {enum_member.value for enum_member in cls}

    @classmethod
    def from_value(cls, value):
        """Create an enum member from the given value, if valid."""
        
        def specific_cleaning(value):
            value = value.replace('actual','act')
            if value == 'act/365fixed':
                value = 'act/365'
            return value        

        cleaned_value = clean_enum_value(value)
        cleaned_value = specific_cleaning(cleaned_value)
        
        if cleaned_value is None:
            return cls.default()        
        for enum_member in cls:
            if enum_member.value == cleaned_value \
                or enum_member.value.replace('/','') == cleaned_value:
                return enum_member

        # List all valid codes in case of an error
        valid_values = [enum_member.value for enum_member in cls]
        raise ValueError(f"Invalid value: {value}. Valid codes are: {valid_values}") 

    @property
    def display_name(self):
        return self.value.upper().replace('_', ' ').strip()


class CompoundingFreq(Enum):
    # If storing instrument definitions in CDM, use 'code' to define valid fieldnames
    SIMPLE = 'simple'
    CONTINUOUS = 'continuous'
    DAILY = 'daily'
    WEEKLY = 'weekly'
    MONTHLY = 'monthly'
    QUARTERLY = 'quarterly'
    SEMIANNUAL = 'semiannual'
    ANNUAL = 'annual'

    def __init__(self, value):        
        periods_per_year_map = {
            'simple': None,
            'continuous': None,
            'daily': 365,
            'weekly': 52,
            'monthly': 12,
            'quarterly': 4,
            'semiannual': 2,
            'annual': 1
            }
        self.periods_per_year = periods_per_year_map[self.value]

    @classmethod
    def is_valid(cls, value):
        return is_valid_enum_value(cls, value)

    @classmethod
    def from_value(cls, value):
        return get_enum_member(cls, value)
        # """Create an enum member from the given value, if valid."""
        # cleaned_value= clean_enum_value(value)
        # for enum_member in cls:
        #     if enum_member.value == cleaned_value:
        #         return enum_member
        # # List all valid codes in case of an error
        # valid_values = [enum_member.value for enum_member in cls]
        # raise ValueError(f"Invalid value: {value}. Valid codes are: {valid_values}")

    @property
    def display_name(self):
        return self.name.title().replace('_', ' ').strip()


class PeriodFreq(Enum):
    # If storing instrument definitions in CDM, use 'code' to define valid fieldnames
    DAILY = 'daily'
    WEEKLY = 'weekly'
    _28_DAYS = '28days'
    MONTHLY = 'monthly'
    QUARTERLY = 'quarterly'
    SEMIANNUAL = 'semiannual'
    ANNUAL = 'annual'
    ZERO_COUPON = 'zerocoupon'
    #CDS = 'cds' # 20th of Mar/Jun/Sep/Dec
    #IMM = 'imm' # 3rd Wednesday of Mar/Jun/Sep/Dec

    def __init__(self, value):        
        date_offset_map = {
            'daily': pd.DateOffset(days=1),
            'weekly': pd.DateOffset(weeks=1),
            '28days': pd.DateOffset(days=28),
            'monthly': pd.DateOffset(months=1),
            'quarterly': pd.DateOffset(months=3),
            'semiannual': pd.DateOffset(months=6),
            'annual': pd.DateOffset(years=1),
            'zerocoupon': None, # No date offset for zero coupon
            #'cds': None, # Custom logic
            #'imm': None, # Custom logic
            }
        self.date_offset = date_offset_map[self.value]

    def multiply_date_offset(self, factor):
        """Multiply the current date_offset by the given factor."""
        # The reason for this function is x * (date += offset) != date += (x * offset).
        # Proof:
        # pd.Timestamp('2023-08-31') + 3 * pd.DateOffset(months=3) == pd.Timestamp('2024-05-29')
        # pd.Timestamp('2023-08-31') + pd.DateOffset(months=3*3) == pd.Timestamp('2024-05-31')
        # The 1st calculation regresses to the shortest end-of-month (Feb)

        if self.date_offset is None:
            return None  # Handle None for ZERO_COUPON case

        return pd.DateOffset(
            days=self.date_offset.kwds.get('days', 0) * factor,
            weeks=self.date_offset.kwds.get('weeks', 0) * factor,
            months=self.date_offset.kwds.get('months', 0) * factor,
            years=self.date_offset.kwds.get('years', 0) * factor
        )

    @classmethod
    def is_valid(cls, value):
        value = clean_enum_value(value, cls._transform_value)
        return value in {enum_member.value for enum_member in cls}

    @classmethod
    def from_value(cls, value):
        return get_enum_member(cls, value, cls._transform_value)

    @staticmethod
    def _transform_value(value: str) -> str:
        value = value.replace('/', '').replace('_', '').replace('per', '').lower().strip()

        period_frequency_aliases = {
            'daily': ['d', 'day', 'days', '1d', '1day', '1days'],
            'weekly': ['w', 'week', 'weeks', 'wk', '1w', '1wk', '1week', '1weeks'],
            '28days': ['28d', '28day', '28days'],
            'monthly': ['m', 'month', 'months', 'mon', '1m', '1mo', '1month', '1months'],
            'quarterly': ['q', 'quarter', 'quarters', 'quart', '3m', '3mo', '3month', '3months'],
            'semiannual': ['s','sa', 'semiannual', 'semiannually', 'semi','6m', '6mo', '6month', '6months'],
            'annual': ['a', 'year', 'years', 'yrs', 'yr', 'annually', '1y', '1yr', '1year', '1years'],
            'zerocoupon': ['z', 'zc', 'zerocoupon', 'zero']
        }

        for key, aliases in period_frequency_aliases.items():
            if value == key or value in aliases:
                return key
        return value

    @property
    def display_name(self):
        return self.name.title().replace('_', ' ').strip()


class DayRoll(Enum):
    UNADJUSTED = 'unadjusted'
    _1 = 1
    _2 = 2
    _3 = 3
    _4 = 4
    _5 = 5
    _6 = 6
    _7 = 7
    _8 = 8
    _9 = 9
    _10 = 10
    _11 = 11
    _12 = 12
    _13 = 13
    _14 = 14
    _15 = 15
    _16 = 16
    _17 = 17
    _18 = 18
    _19 = 19
    _20 = 20
    _21 = 21
    _22 = 22
    _23 = 23
    _24 = 24
    _25 = 25
    _26 = 26
    _27 = 27
    _28 = 28
    _29 = 29
    _30 = 30
    _31 = 31
    EOM = 'eom' # End of month

    @classmethod
    def default(cls):
        return cls.UNADJUSTED # Return the default enum value

    @classmethod
    def is_valid(cls, value):
        return is_valid_enum_value(cls, value)

    @classmethod
    def from_value(cls, value):
        return get_enum_member(cls, value)

    @property
    def display_name(self):
        return self.name.title().replace('_', ' ').strip()
        
        
    
class RollConv(Enum):
    UNADJUSTED = 'unadjusted'
    FOLLOWING = 'following'
    PRECEDING = 'preceding'
    MODIFIED_FOLLOWING = 'modifiedfollowing'
    MODIFIED_PRECEDING = 'modifiedpreceding'

    @classmethod
    def is_valid(cls, value):
        return is_valid_enum_value(cls, value)

    @classmethod
    def from_value(cls, value):
        return get_enum_member(cls, value)

    @classmethod
    def default(cls):
        return cls.MODIFIED_FOLLOWING # Return the default enum value

    @property
    def display_name(self):
        return self.name.replace('_', ' ').title()


class TimingConvention(Enum):
    IN_ARREARS = 'in_arrears'
    IN_ADVANCE = 'in_advance'

    @classmethod
    def is_valid(cls, value):
        return is_valid_enum_value(cls, value)

    @classmethod
    def from_value(cls, value):
        return get_enum_member(cls, value)
        
    @property
    def display_name(self):
        return self.name.title().replace('_', ' ')

    @classmethod
    def fixing_default(cls):
        return cls.IN_ADVANCE # Return the default enum value

    @classmethod
    def payment_default(cls):
        return cls.IN_ARREARS # Return the default enum value

class Stub(Enum):
    NONE = 'none'
    SHORT = 'short'
    LONG = 'long'
    DEFAULT = 'default'
    DEFINED_PER_FIRST_PERIOD_END_DATE = 'defined_per_first_period_end_date'
    DEFINED_PER_LAST_PERIOD_START_DATE = 'defined_per_last_period_start_date'

    @classmethod
    def is_valid(cls, value):
        return is_valid_enum_value(cls, value)

    @classmethod
    def from_value(cls, value):
        return get_enum_member(cls, value)

    @classmethod
    def default(cls):
        return cls.DEFAULT  # Return the default enum value

    @classmethod
    def market_convention(cls):
        return cls.SHORT  # Market convention is a short stub

    @property
    def display_name(self):
        return self.name.title().replace('_', ' ')
        

class ExchangeNotionals(Enum):
    START = 'start'
    END = 'end'
    BOTH = 'both'
    NEITHER = 'neither'

    @classmethod
    def is_valid(cls, value):
        return is_valid_enum_value(cls, value)

    @classmethod
    def from_value(cls, value):
        return get_enum_member(cls, value)

    @property
    def display_name(self):
        return self.name.title()

    @classmethod
    def default(cls):
        return cls.NEITHER # Return the default enum value


